#!/usr/bin/python

# Update Jan 3, 2012 by Natalie to remove dependence on Column Order

UsageInfo = \
"""InspectToPepXML.py: Converts output of InsPecT search engine
to PepXML format.  Written by Samuel Payne, Venter Institute,
and Terry Farrah, Institute for Systems Biology  October 2008

Required Parameters
-i [Filename] - InsPecT results file from search (input)
-o [Filename] - Converted file in PepXML (output)

Optional Parameters
-p [Filename] - InsPecT input (parameter) file
                  default: inspect.params
-m [Dirname] - Dir containing .mzXML or .mgf spectrum file
                  default: current working directory
-d N   - write at most N hits per assumed charge

Assumes InsPecT results file is TSV containing header line and
 one record per peptide prediction sorted by scan #, then by rank.
User must manually edit PepXML file and insert correct information
  near top of file for precursor and fragment mass types --
  either average or monoisotopic.
If database file mentioned in parameter file is not in fasta
  format (.fasta or .fa), you must create a fasta format file of
  the same base name in the same dir. Use TrieToFASTA.py.
This script, InspectToPepXML.py, must reside in the same directory
  as the rest of the InsPecT code.
"""

import sys
import os
import glob
import getopt
import re
import time
import GetByteOffset
import ResultsParser
import Utils
import Global
from xml.sax import saxutils          #xml.sax is for reading mzXML
from xml.sax import ContentHandler
from xml.sax import make_parser
from xml.sax.handler import feature_namespaces

global initial_dir
global spectrum_query_count

# ========================================================
# Read tables with standard data such as amino acid masses
# ========================================================

# chdir() is Hack to make pgm invokable from any dir
#  (Utils makes use of auxiliary files in same dir as code)
initial_dir = os.getcwd()
os.chdir(sys.path[0])
Utils.Initialize()  
os.chdir(initial_dir)

# ===========================================================
# Define classes to hold spectra (scans) and peptides (hits)
# ===========================================================

class InspectSpectrumClass:
    """Stores the relevant InsPecT output file data for a spectrum"""
    def __init__(self):
        self.ScanNumber = -1
        self.PrecursorMz = -1.0
        self.RetentionTime = -1.0
        self.HitList = []   # store a hit list for each charge state
        for i in range (1,6): self.HitList = self.HitList + [[]]

    def WriteSpectrumQueries(self, PepXMLHandle, SpectrumFileName, enzyme,
            MaxHitsPerCharge):
        """ Write <spectrum_query> tags for this spectrum.

            There is one tag for each assumed charge that has any hits.
        """
        global spectrum_query_count
        SpectrumFileType = os.path.splitext(SpectrumFileName)[1].lower()

        for charge in range(1,5):     # for each charge state
            if len(self.HitList[charge]) > 0: # if any hits
                spectrum_query_count = spectrum_query_count + 1
                SpectrumTitle="%s.%05d.%05d.%s" % \
                   (os.path.splitext(SpectrumFileName)[0],
                     self.ScanNumber,self.ScanNumber,
                     charge)
                if SpectrumFileType == ".mgf":
                  PrecursorNeutralMass = self.PrecursorNeutralMass
                else:
                  _proton_mass = 1.007276
                  PrecursorNeutralMass =  \
                      (self.PrecursorMz * charge) - \
                      (charge * _proton_mass)
                if PrecursorNeutralMass < 0:
                  PrecursorNeutralMassString = ''
                else:
                  PrecursorNeutralMassString = \
                    '    precursor_neutral_mass="%.6f"\n' % \
                                  PrecursorNeutralMass
                if self.RetentionTime < 0:
                  RetentionTimeString = ''
                else:
                  RetentionTimeString = \
                    '    retention_time_sec="%.2f"\n' % \
                                  self.RetentionTime
                Query = '<spectrum_query\n' + \
                    '    spectrum="%s"\n' % SpectrumTitle + \
                    '    start_scan="%s"\n' % self.ScanNumber + \
                    '    end_scan="%s"\n' % self.ScanNumber + \
                    PrecursorNeutralMassString + \
                    '    assumed_charge="%s"\n' % charge + \
                    '    index="%s"\n' % spectrum_query_count + \
                    RetentionTimeString + \
                    '>\n'
                PepXMLHandle.write(Query)
                PepXMLHandle.write('<search_result search_id="1">\n')
                for i in range(min(MaxHitsPerCharge,
                         len(self.HitList[charge]))):
                    self.HitList[charge][i].PrecursorNeutralMass = \
                          PrecursorNeutralMass
                    self.HitList[charge][i].WriteSearchHit(PepXMLHandle,
                          i+1, enzyme)
                PepXMLHandle.write('</search_result>\n')
                PepXMLHandle.write('</spectrum_query>\n')

class InspectOutputRecordClass:
    """Stores the relevant data from a single line of InsPecT output.

    Each line represents a search hit--a predicted peptide for a spectrum.
    """
    def __init__(self):
        self.Spectrum = None
        self.FileOffset = -1
        self.Protein = ""
        self.Charge = -1
        self.MQScore = ""
        self.FScore = ""
        self.DeltaScore = ""
        self.PValue = ""
        self.ProteinID = ""
        self.Prefix = ""
        self.Peptide = ""
        self.Suffix = ""
        self.OptModList = []
        self.PrecursorNeutralMass = -1.0

    def WriteSearchHit(self, PepXMLHandle, rank, enzyme):
        """ Write <search_hit> tag for this this line of InsPecT output
        """           
        global initial_dir
        os.chdir(sys.path[0]) # hack to make pgm invokable from any dir
        # GetMass adds on fixed modifications, but not optional ones
        CalcMass = Utils.GetMass(self.Peptide) + 18.01528 #add h2o mass
        for mod in self.OptModList:
            CalcMass = CalcMass + float(mod[2])
        os.chdir(initial_dir)
        MassDiff = self.PrecursorNeutralMass - CalcMass
        # If the enzyme is trypsin, count all KR except
        # final one, and except when followed by P (proline).
        if enzyme.lower() == "trypsin":
            MissedCleavages = self.Peptide[:-1].count("K") + \
               self.Peptide[:-1].count("R") - \
               (self.Peptide[:-1].count("KP") + self.Peptide[:-1].count("RP"))
        elif enzyme.lower() == "none":
            MissedCleavages = 0
        else: MissedCleavages = -1
        # Break up Protein into accession # and description
        first_space = self.Protein.find(' ')
        if first_space >= 0:
          Protein = self.Protein[:first_space]
          ProteinDescr = self.Protein[first_space+1:]
          ProteinDescr = ProteinDescr.replace(">","&gt;")
          ProteinDescr = ProteinDescr.replace("<","&lt;")
          ProteinDescr = ProteinDescr.replace("&","&amp;")
          ProteinDescr = ProteinDescr.replace("\"","&quot;")
          ProteinDescr = ProteinDescr.replace("\'","&lsquo;")
        else:
          Protein = self.Protein
          ProteinDescr = Protein
        Hit = '<search_hit\n' + \
              '    hit_rank="%s"\n' % (rank) + \
              '    peptide="%s"\n' % (self.Peptide) + \
              '    peptide_prev_aa="%s"\n' % (self.Prefix) + \
              '    peptide_next_aa="%s"\n' % (self.Suffix) + \
              '    protein="%s"\n' % (Protein) + \
              '    protein_descr="%s"\n' % (ProteinDescr) + \
              '    num_tot_proteins="0"\n' + \
              '    num_matched_ions="0"\n'  + \
              '    tot_num_ions="0"\n' + \
              '    calc_neutral_pep_mass="%s"\n' % (CalcMass) + \
              '    massdiff="%s"\n' % (MassDiff) + \
              '    num_tol_term="%s"\n' % "2" + \
              '    num_missed_cleavages="%d"\n'%(MissedCleavages) + \
              '    is_rejected="0"\n' + \
              '>\n'
        PepXMLHandle.write(Hit)
        # Create a dictionary of masses of all amino acids that
        # are modified, indexed by peptide position.
        # First, add to the dictionary all aa's that have optional mods.
        # Then, add fixed mods. Use monoisotopic mass for basic AA.
        ModMassDict = {}
        for mod in self.OptModList:
           aa = mod[0]
           pos = mod[1]
           mod_mass = mod[2]
           if pos in ModMassDict:
             ModMassDict[pos] += float(mod_mass)
           else:
             ModMassDict[pos] = float(mod_mass) + Global.AminoMass[aa]
        for i in range(len(self.Peptide)):
           aa = self.Peptide[i]
           pos = i + 1
           if aa in Global.FixedMods:
               mod_mass = Global.FixedMods[aa]
               if pos in ModMassDict:
                   ModMassDict[pos] += float(mod_mass)
               else:
                   ModMassDict[pos] = float(mod_mass) + \
                                            Global.AminoMass[aa]
        # Now, create a pepXML string with an element for each modified AA.
        ModString = ''
        for i in range(len(self.Peptide)):
            pos = i + 1
            if pos in ModMassDict:
                ModString = ModString + '<mod_aminoacid_mass ' + \
                  'position="%d" ' % pos + \
                  'mass="%.4f" />' % ModMassDict[pos]
        if len(ModString) > 0:
            ModInfo = '<modification_info>%s</modification_info>\n' % \
                   ModString
            PepXMLHandle.write(ModInfo)
        PepXMLHandle.write(
          '<search_score name="mqscore" value="%s"/>\n'%self.MQScore)
        PepXMLHandle.write(
          '<search_score name="expect" value="%s"/>\n'%self.PValue)
        PepXMLHandle.write(
          '<search_score name="fscore" value="%s"/>\n'%self.FScore)
        PepXMLHandle.write(
          '<search_score name="deltascore" value="%s"/>\n'%self.DeltaScore)
        PepXMLHandle.write('</search_hit>\n')


# ======================================================================
# Virtually all the code is contained within class InspectToPepXMLClass
# ======================================================================

class InspectToPepXMLClass(ResultsParser.ResultsParser):
    def __init__(self):
        """Initialize fields of InspectToPepXMLClass instance to null values
        """
        self.InputFilePath = None
        self.OutputFilePath = None
        self.SpectraDir = os.getcwd()
        self.MaxHitsPerCharge = 10000   #effectively maxint
        self.ParamFilePath = os.path.join(os.getcwd(), "inspect.params")
        self.ScanOffset = {}
        self.ScanDict= {}
        self.SpectrumFileType = ""
        self.SpectrumFileBase = ""
        self.Columns = ResultsParser.Columns()
        ResultsParser.ResultsParser.__init__(self)
        
    #---------------------------------------------------------------------

    def Main(self):
        """Convert raw InsPecT output file to PepXML

        Initially designed to handle entire directories of files.
        """
        try:
            import psyco
            psyco.full()
        except:
            print "(psyco not found - running in non-optimized mode)"
        # Line directly below needed only if we want to handle directories
        #self.self.ProcessResultsFiles(self.InputFilePath,
        #    self.ConvertInspectToPepXML)
        self.ConvertInspectToPepXML(self.InputFilePath)

    #---------------------------------------------------------------------

    def ConvertInspectToPepXML(self, FilePath):
        """ Convert a single raw InsPecT output file to PepXML
        """

        global spectrum_query_count

        # ------------------------------------------------------------
        # Open input/output files and gather info from auxiliary files
        # ------------------------------------------------------------

        # Get input filename; open output file handle
        #FileName = os.path.split(FilePath)[1]
        #FileName = FileName.replace(".txt", ".xml")
        #NewPath = os.path.join(self.OutputFilePath, FileName)
        #PepXMLHandle = open(NewPath, "wb")
        PepXMLHandle = open(self.OutputFilePath, "wb")

        # Glean info from inspect params file
        if not os.path.exists(self.ParamFilePath):
            print >> sys.stderr, "Inspect params file %s does not exist" % \
                  self.ParamFilePath
            sys.exit()
        ParamFile = open(self.ParamFilePath, "r")
        nmods_allowed_per_spectrum = 0
        nmods_in_params = 0
        self.mod_weight = []
        self.mod_aa = []
        self.mod_type = []
        self.mod_name = []
        self.spec_file = []
        # reset Global.FixedMods to empty; Global.py initializes it to
        # {"C":57.0518}, but this is a hack we don't want
        Global.FixedMods = {}
        self.instrument = "UNKNOWN"
        self.protease = "trypsin"
        self.search_db = "UNKNOWN"
        for Line in ParamFile.readlines():
           Line = Line.strip()  #remove leading and trailing whitespace
           if Line.lower().startswith("mods,"):
               nmods_allowed_per_spectrum = int(Line[len("mods,"):])
           elif Line.lower().startswith("spectra,"):
               this_spec_file =  Line[len("spectra,"):].strip()
               self.spec_file = self.spec_file + [this_spec_file]
           elif Line.lower().startswith("mod,"):
               tokens = Line.split(",")
               this_mod_weight =  float(tokens[1])
               this_mod_aa_string = tokens[2].strip()
               if this_mod_aa_string == "*":
                 this_mod_aa_string = "ACDEFGHIKLMNPQRSTVWY"
               this_mod_type = "opt"
               if len(tokens) > 3:
                   this_mod_type = tokens[3].strip()
               if len(tokens) > 4:
                   this_mod_name = tokens[4].strip()
               else: this_mod_name = None
               for this_mod_aa in this_mod_aa_string:
                   self.mod_weight = self.mod_weight + [this_mod_weight]
                   self.mod_aa = self.mod_aa + [this_mod_aa]
                   self.mod_type = self.mod_type + [this_mod_type]
                   if this_mod_type == "fix":
                       Global.FixedMods[this_mod_aa] = this_mod_weight
                   self.mod_name = self.mod_name + [this_mod_name]
                   nmods_in_params = nmods_in_params + 1
           elif Line.lower().startswith("instrument,"):
               self.instrument = Line[len("instrument,"):].strip()
           elif Line.lower().startswith("protease,"):
               self.protease = Line[len("protease,"):].strip()
           elif Line.lower().startswith("db,"):
               search_db = Line[len("db,"):].strip()
               search_db_ext = os.path.splitext(search_db)[1]
               # Find the Fasta format of the .trie file used
               if search_db_ext not in [".fa", ".fasta"]:
                   search_db_root = os.path.splitext(search_db)[0]
                   search_db_file_list = set(
                           glob.glob("%s.*" % search_db_root))
                   #print search_db_file_list
                   ext_list = [os.path.splitext(f)[1]
                           for f in search_db_file_list]
                   #print ext_list
                   try: ext_list.remove(".index")
                   except: pass
                   try: ext_list.remove(".trie")
                   except: pass
                   #print ext_list
                   if len(ext_list) == 1:
                       search_db = search_db_root + ext_list[0]
                   elif ".fasta" in ext_list:
                       search_db = search_db_root + ".fasta"
                   elif ".fa" in ext_list:
                       search_db = search_db_root + ".fa"
                   else:
                       print >> sys.stderr, \
           "WARNING: Can't find a RefreshParser compatible database " + \
           "file corresponding to %s " % search_db + \
           "(such as a .fasta or .fa file with same root); using UNKNOWN.\n" + \
           "(%s is the database file listed in your params file.)\n" % search_db
               self.search_db = search_db
        self.nmods = nmods_in_params

        # Read just first line of inspect output to get spectrum filename
#        InspectHandle = open(FilePath, "r")
#        for Line in InspectHandle.xreadlines():
#            if Line[0] == "#":  # comments
#                continue
#            Bits = list(Line.split("\t"))
#            break
#        InspectHandle.close()
        InspectHandle = open(FilePath, "rb")

        # Glean RTs & precursor M/z's for each scan from each spectrum file
        # Also, store the full path for each file in a dictionary
        # keyed to the filename.
        retentionTimeDict = dict()
        precursorMzDict = dict()
        spectrumPathDict = dict()      ###TMF_new
        for SpectrumFilePath in self.spec_file:
#	  SpectrumFilePath = Bits[self.Columns.SpectrumFile]
	  SpectrumFileName = os.path.split(SpectrumFilePath)[1]
# We used to force the path to be the cwd, but now we're leaving it alone. Dec-11
#	  SpectrumFilePath = os.path.join(self.SpectraDir, SpectrumFileName)
	  self.SpectrumFileBase = \
	    SpectrumFilePath.replace(os.path.splitext(SpectrumFilePath)[1], "")
	  #self.SpectrumFileType = os.path.splitext(SpectrumFilePath)[1]
          spectrumPathDict[SpectrumFileName] = SpectrumFilePath

	  if not os.path.exists(SpectrumFilePath):
	      print >> sys.stderr, "Spectrum file %s does not exist" % \
		    SpectrumFilePath
	      sys.exit()
	  (Stub, Ext) = os.path.splitext(SpectrumFilePath)
	  if Ext.lower() == ".mzxml":
	      self.SpectrumFileType = ".mzXML"
	      (this_retentionTimeDict, this_precursorMzDict) =  \
		  self.GetSpectrumInfoFromMzXML(SpectrumFilePath)
              retentionTimeDict.update(this_retentionTimeDict)
              precursorMzDict.update(this_precursorMzDict)
	  elif Ext.lower() == ".mgf":
	      self.SpectrumFileType = ".mgf"
              break
	  else:
	      print >> sys.stderr, \
		 "Spectrum file %s lacks .mzXML or .mgf extension" % \
		    SpectrumFilePath
	      sys.exit()

        # ------------------------------------------------------------
        # - Write opening info to PepXML
        # - Process InsPecT output file line by line and write to PepXML
        # - Write closing info to PepXML
        # ------------------------------------------------------------

        self.WritePepXMLOpening(PepXMLHandle, self.OutputFilePath)

        LastScanNumber = -1
        spectrum_query_count = 0

        # Each line represents a predicted peptide for a spectrum (scan).
        # A scan can have multiple predicted peptides (hits).
        #  All hits for a scan are grouped together in the file.
        #  Further, all scans for each spectrum file are grouped
        #  together.
        for Line in InspectHandle.xreadlines():
            if Line[0] == "#": 
                self.Columns.initializeHeaders(Line) #This is the header, so save it
                continue  # skip comments
            # create a record for this line and read the fields into Bits
            this_rec = InspectOutputRecordClass()
            Bits = list(Line.split("\t"))

            try:
              this_rec.FileOffset = int(Bits[self.Columns.getIndex("SpecFilePos")])
            except:
              print "WARNING: malformed FileOffset %s in/after scan %d" % (Bits[self.Columns.getIndex("SpecFilePos")], LastScanNumber)
              continue

            try:
              ScanNumber = int(Bits[self.Columns.getIndex("Scan#")])
            except:
              print "WARNING: malformed ScanNumber %s in/after scan %d" % (Bits[self.Columns.getIndex("Scan#")], LastScanNumber)
              continue

            ### TMF_new
            try:
              SpectrumFilePath = Bits[self.Columns.getIndex("SpectrumFile")]
              SpectrumFile = os.path.split(SpectrumFilePath)[1]
            except:
              print "WARNING: malformed SpectrumFile field in/after scan %d" % (LastScanNumber)
              continue

            ScanName = SpectrumFile + "." + str(ScanNumber)

            if (LastScanNumber != ScanNumber): 
                if (LastScanNumber != -1):
                    # write results for last spectrum
                    this_scan.WriteSpectrumQueries(PepXMLHandle,
#                        SpectrumFileName, self.protease,
                        SpectrumFile, self.protease,
                        self.MaxHitsPerCharge)
                # initialize new spectrum
                this_scan = InspectSpectrumClass()
                this_scan.ScanNumber = ScanNumber
                this_scan.ScanName = ScanName
                # get info about spectrum from spectrum file
                if self.SpectrumFileType == ".mgf":
                    SpectrumFilePath = spectrumPathDict[SpectrumFile] ### TMF_new
                    (MgfPepMass, MgfRT) = \
                       self.GetSpectrumInfoFromMGF(SpectrumFilePath,
                       this_rec.FileOffset)
                    this_scan.PrecursorNeutralMass = float(MgfPepMass)
                    this_scan.RetentionTime = float(MgfRT)
                elif self.SpectrumFileType == ".mzXML":
                  if not retentionTimeDict.has_key(this_scan.ScanName):
                    print "WARNING: RT for scan %s not found in spectrum file; retention_time_sec will not be output" % ScanName
                    this_scan.RetentionTime = -1.0
                  else:
                    this_scan.RetentionTime = \
                          retentionTimeDict[this_scan.ScanName]
                  if not precursorMzDict.has_key(this_scan.ScanName):
                    print "WARNING: m/z for scan %s not found in spectrum file; precursor_neutral_mass will not be output" % ScanName
                    this_scan.PrecursorMz = -1.0
                  else:
                    this_scan.PrecursorMz = \
                           precursorMzDict[this_scan.ScanName]

            this_rec.Spectrum = this_scan
            LastScanNumber = ScanNumber

            # ---------------------------
            # Process data about this hit
            # ---------------------------
            Annotation = Bits[self.Columns.getIndex("Annotation")]
            Peptide = Annotation[2:-2]

            # process peptide string --TMF
            # I think there is already code to do this in Utils.py
            # Sam, you may want to replace my code with a call to that.
            def ExtractAAModifications(search, peptide):
              '''Given peptide like TVAM+16GGKYphosLV, extract the numbers
                 and other modification symbols.

                 Return (a) the peptide without the mods, and
                 (b) a list of (aa, aa-pos, number) tuples -- 
                 aa/aa-pos describe the aa posessing the mod.
              '''
              i = 0
              mod_list = []
              stripped_peptide = ""
              while i < len(peptide):
                if peptide[i].isupper():
                  stripped_peptide = stripped_peptide + peptide[i]
                  i = i + 1
                  continue
                j = i + 1
                while j < len(peptide) and not peptide[j].isupper():
                  j = j + 1
                aa = peptide[i-1]
                added_mod = peptide[i:j]
                added_mod_pos = len(stripped_peptide) #counting starts at 1
                # modifications with names in the param file
                # will be represented by their names embedded in the
                # peptide. Look up their weights.
                for k in range(nmods_in_params):
                  if search.mod_name[k]:
                     truncated_name = search.mod_name[k][:4]
                     this_weight = int(search.mod_weight[k])
                     if this_weight > 0:
                       weight_string = "+" + str(this_weight)
                     added_mod = added_mod.replace(truncated_name,
                         weight_string)
                # added_mod could be a concatenation of several mods,
                #  as in AEQDNLGKSVM-5+16IPTK;
                # store each one as a separate mod.
                this_mod = ""
                for i in range(len(added_mod)):
                  c = added_mod[i]
                  if (c == "+" or c == "-"):
                    # store the previous mod
                    if len(this_mod) > 0:
                      mod_list = mod_list + [(aa, added_mod_pos, this_mod)]
                    # start a new mod
                    this_mod = c
                  else:
                    this_mod = this_mod + c
                # store the last mod
                mod_list = mod_list + [(aa, added_mod_pos, this_mod)]
                i = j
              return (stripped_peptide, mod_list)

            (this_rec.Peptide, this_rec.OptModList) = \
                 ExtractAAModifications(self, Peptide)

            # done processing peptide string
   
            this_rec.Prefix = Annotation[0]
            this_rec.Suffix = Annotation[-1]
            this_rec.Protein = Bits[self.Columns.getIndex("Protein")]
            this_rec.Charge = int(Bits[self.Columns.getIndex("Charge")])
            this_rec.MQScore = Bits[self.Columns.getIndex("MQScore")]
            this_rec.FScore = Bits[self.Columns.getIndex("F-Score")]
            this_rec.DeltaScore = Bits[self.Columns.getIndex("DeltaScore")]
            this_rec.PValue = Bits[self.Columns.getIndex("InspectFDR")]
            this_rec.ProteinID = Bits[self.Columns.getIndex("RecordNumber")]

            this_scan.HitList[this_rec.Charge] = \
                this_scan.HitList[this_rec.Charge] + [this_rec]
             
            # done processing a single line of InsPecT output file

        # write conversion of last line of InsPecT output file
        this_scan.WriteSpectrumQueries(PepXMLHandle, \
#                       SpectrumFileName, self.protease, \
                       SpectrumFile, self.protease, \
                       self.MaxHitsPerCharge)

        self.WritePepXMLClosing(PepXMLHandle)
        InspectHandle.close()
        PepXMLHandle.close()


    #---------------------------------------------------------------------

    def WritePepXMLOpening(self, PepXMLHandle, PepXMLFilePath):
        """Write stuff that belongs at the top of the pepXML file"""
        PepXMLHandle.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        PepXMLHandle.write('<?xml-stylesheet type="text/xsl" href="pepXML_std.xsl"?>\n')
        datestr = time.strftime('%Y-%m-%dT%H:%M:%S')
        PepXMLHandle.write(
           '<msms_pipeline_analysis ' + 
             'date="%s" ' % datestr  +
             'summary_xml="%s" ' %PepXMLFilePath +
             'xmlns="http://regis-web.systemsbiology.net/pepXML" ' +
             'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" ' +
             'xsi:schemaLocation="http://regis-web.systemsbiology.net/pepXML /tools/bin/TPP/tpp/schema/pepXML_v112.xsd" ' +
           '>\n' )
           #'xsi:schemaLocation="http://regis-web.systemsbiology.net/pepXML http://mascot1/mascot/xmlns/schema/pepXML_v18/pepXML_v18.xsd" ' +
        PepXMLHandle.write(
           '<msms_run_summary ' +
             'base_name="%s" ' % self.SpectrumFileBase +
             'search_engine="InsPecT" ' +
             'msManufacturer="UNKNOWN" ' +
             'msModel="%s" ' % self.instrument +
             'msIonization="UNKNOWN" ' +
             'msMassAnalyzer="UNKNOWN" ' +
             'msDetector="UNKNOWN" ' +
             'raw_data_type="raw" ' +
             'raw_data="%s" ' % self.SpectrumFileType +
           '>\n')
        PepXMLHandle.write('<sample_enzyme name="%s">\n' % self.protease)
        PepXMLHandle.write('<specificity cut="KR" no_cut="P" sense="C"/>\n')
        PepXMLHandle.write('</sample_enzyme>\n')
        PepXMLHandle.write(
           '<search_summary ' +
             'base_name="%s" ' % self.SpectrumFileBase +
             'search_engine="InsPecT" ' +
             'precursor_mass_type="monoisotopic" ' +
             'fragment_mass_type="monoisotopic" ' +
             'search_id="1" ' +
             'out_data_type="out" ' +
             'out_data=".txt" ' +
           '>\n')
        PepXMLHandle.write(
           '<search_database ' +
             'local_path="%s" ' % self.search_db +
             'type="AA" ' +
           '/>\n')
        #'database_name="" ' +
        #'database_release_identifier="" ' +
        #'size_in_db_entries="" ' +
        #'size_of_residues="" ' +
        PepXMLHandle.write(
           '<enzymatic_search_constraint ' +
              'enzyme="%s" ' % self.protease +
              'max_num_internal_cleavages="2" ' +
              'min_number_termini="2" ' +
           '/>\n')
        for i in range(self.nmods):
            mod_aa = self.mod_aa[i]
            mod_weight = self.mod_weight[i]
            mass = mod_weight + Global.AminoMass[mod_aa]
            if self.mod_type[i] == "opt": mod_variable="Y"
            elif self.mod_type[i]=="fix": mod_variable="N"
            else: mod_variable="UNKNOWN"   # are there other types?
            PepXMLHandle.write(
               '<aminoacid_modification ' +
                'aminoacid="%s" ' % mod_aa +
                'massdiff="%.4f" ' % mod_weight +
                'mass="%.4f" ' % mass +
                'variable="%s" ' % mod_variable +
               '/>\n')
        PepXMLHandle.write('<parameter name="CHARGE" value="2+ and 3+"/>\n')
        PepXMLHandle.write('<parameter name="CLE" value="Trypsin"/>\n')
        PepXMLHandle.write('<parameter name="DB" value=""/>\n')
        PepXMLHandle.write('<parameter name="FILE" value=""/>\n')
        PepXMLHandle.write('<parameter name="FORMAT" value=""/>\n')
        PepXMLHandle.write('<parameter name="FORMVER" value=""/>\n')
        PepXMLHandle.write('<parameter name="INSTRUMENT" value="%s"/>\n' % \
          self.instrument)
        PepXMLHandle.write('<parameter name="ITOL" value=""/>\n')
        PepXMLHandle.write('<parameter name="ITOLU" value="Da"/>\n')
        PepXMLHandle.write('<parameter name="MASS" value="Monoisotopic"/>\n')
        PepXMLHandle.write('<parameter name="REPORT" value=""/>\n')
        PepXMLHandle.write('<parameter name="REPTYPE" value="Peptide"/>\n')
        PepXMLHandle.write('<parameter name="RULES" value=""/>\n')
        PepXMLHandle.write('<parameter name="SEARCH" value=""/>\n')
        PepXMLHandle.write('<parameter name="TAXONOMY" value=""/>\n')
        PepXMLHandle.write('<parameter name="TOL" value=""/>\n')
        PepXMLHandle.write('<parameter name="TOLU" value="Da"/>\n')
        PepXMLHandle.write('</search_summary>\n')

    #---------------------------------------------------------------------

    def WritePepXMLClosing(self, PepXMLHandle):
        """Write stuff that belongs at the end of the pepXML file"""
        PepXMLHandle.write('</msms_run_summary>\n')
        PepXMLHandle.write('</msms_pipeline_analysis>\n')

    #---------------------------------------------------------------------

    def GetAllSpectrumInfoFromMGF(self, FilePath):
      sys.exit(1)

    #---------------------------------------------------------------------

    def GetSpectrumInfoFromMGF(self, FilePath, FileOffset):
        """ returns the spectrum title and peptide mass corresponding to
            the spectrum at the given file offset in the given mgf file
        """
        File = open(FilePath, "r")

        File.seek(FileOffset)
        Mass = 0
        RT = 0
        Title = None

        MatchMass = re.compile('^PEPMASS=(\S*)')
        MatchRT = re.compile('^RTINSECONDS=(\S*)')
        MatchTitle = re.compile('^TITLE=([^\n]*)')
        # read one line at a time
        for Line in File:
           
            # We are not currently using the title
            #Match = MatchTitle.match(Line)
            #if Match != None:   
                #Title = Match.group(1)
                #continue         

            # is this a mass line?
            Match = MatchMass.match(Line)
            if Match != None:   
                Mass = Match.group(1)
                continue         

            # is this an RT line?
            Match = MatchRT.match(Line)
            if Match != None:   
                RT = Match.group(1)
                continue

            # this is not title, mass, charge, or RT. If we've read
            # all of them already, stop reading.
            if  Mass!=0 and RT!=0:
                break
           
        File.close()
        if  Mass==0 or  RT==0:
          print >> sys.stderr, "WARNING: mass, and/or RT missing for spectrum at offset %s in %s" % ( FileOffset, FilePath )
        return (Mass,RT)

    #--------------------------------------------------------------------

    def GetSpectrumInfoFromMzXML(self, FilePath):
        """ compiles dictionaries of the precursorMz and retentionTime
            for each spectrum in an mzXML file
        """

	def normalize_whitespace(text):
	    "Remove redundant whitespace from a string"
	    return ' '.join(text.split())

	class MzXMLHandler(ContentHandler):

	    def __init__(self):
                self.this_scan = None
                self.this_precursorMz = None
                self.precursorMz = dict()
                self.retentionTime = dict()
                self.inPrecursorMzContent = 0
                self.FileName = os.path.split(FilePath)[1]  ###TMF_new

	    def startElement(self, name, attrs):
		# If it's not a comic element, ignore it
		if name == 'scan':
		    # Look for the title and number attributes
		    num = int(normalize_whitespace(attrs.get('num', None)))
		    retentionTime = normalize_whitespace(
				     attrs.get('retentionTime', None))
		    self.this_scan = int(num)
                    self.this_scan_name = self.FileName + "." + str(num) ###TMF_new
		    self.retentionTime[self.this_scan_name] = \
                        float(retentionTime[2:-1])
#		    self.retentionTime[self.this_scan] = \
#                        float(retentionTime[2:-1])
		elif name == 'precursorMz':
		    self.inPrecursorMzContent = 1
		    self.thisprecursorMz = ""

	    def characters(self, ch):
		if self.inPrecursorMzContent:
		    self.thisprecursorMz = self.thisprecursorMz + ch

	    def endElement(self, name):
		if name == 'precursorMz':
		    self.inPrecursorMzContent = 0
                    idx = self.this_scan_name  ###TMF_new
                    self.precursorMz[idx] = float(self.thisprecursorMz)
#                    i = self.this_scan
#                    self.precursorMz[i] = float(self.thisprecursorMz)
		elif name == 'scan':
                    pass

	# Create an XML parser and tell it
	# we are not interested in XML namespaces
	MzXMLparser = make_parser()
	MzXMLparser.setFeature(feature_namespaces, 0)

	# Create a handler and tell the parser to use it
	mh = MzXMLHandler()
	MzXMLparser.setContentHandler(mh)

	# Parse the file
        File = open(FilePath, "r")
        try:
	  MzXMLparser.parse(File)
        except:
          print >> sys.stderr, "ERROR: SAX parser cannot parse %s" % FilePath
          sys.exit()

        return (mh.retentionTime, mh.precursorMz)

    #---------------------------------------------------------------------

    def ParseCommandLine(self,Arguments):
        (Options, Args) = getopt.getopt(Arguments, "i:o:m:p:d:")
        OptionsSeen = {}
        for (Option, Value) in Options:
            OptionsSeen[Option] = 1
            if Option == "-i":
                if not os.path.exists(Value):
                  print "** Error: couldn't find results file '%s'\n\n"%Value
                  print UsageInfo
                  sys.exit(1)
                self.InputFilePath = Value
            if Option == "-o":
                self.OutputFilePath = Value
            if Option == "-m":
                self.SpectraDir = Value
            if Option == "-p":
                self.ParamFilePath = Value
            if Option == "-d":
                self.MaxHitsPerCharge = int(Value)
        if not OptionsSeen.has_key("-i") or not OptionsSeen.has_key("-o"):
            print UsageInfo
            sys.exit(1)

    def Finish(self):   
        self.InputFile.close()
        self.OutputFile.close()

#-------------------------------------------------------------------------

if __name__ == '__main__':
    Fix = InspectToPepXMLClass()
    Fix.ParseCommandLine(sys.argv[1:])  
    Fix.Main()



#Title:          PhosphateLocalization.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.


"""PhosphateLocalization.py

This script is a glorified wrapper for Label.py. It calls label and
calculates the PLS score for each spectral annotation in the input set.

1. read in input data.  If it is not in native inspect format, then we
send it to the GetByteOffset part, so that we can use it like Inspect.

2. Label all possible annotations of the string and get their peptide score (think binomial)

3. Find the difference between the top two scores, report. print.
The results are reported by appending two extra columns to the data from the input
file.  These correspond to the top annotation, and it's PLS.
"""

UsageInfo = """PhosphateLocalization.py
Calculates the Phosphate Localization Score (PLS) for each spectral
annotation in the input file.  Make sure to read the tutorial so
that you understand how to use it correctly.

Required Options:
 -r [FileName] File of formatted annotations
 -m [Directory] Directory containing spectra files (not filename)
 -w [FileName] Output of this program

Additional Options:
 -d [Directory] Directory for the images and annotated peak lists
      created during the label process.  Default "LabelSpewage"

"""

import os
import sys
import getopt
import ResultsParser
import GetByteOffset
import string
import Label

class DetectiveClass(ResultsParser.ResultsParser):
    def __init__(self):
        self.InputFilePath = None
        self.OutputFilePath = None
        self.LabeledAnnotationsDir = "LabelSpewage" # for Labeled output
        self.MZXMLDir = None
        self.InspectFormat = 0
        self.ScanOffset = {} # potentially large dictionary for storing the byte offset of each spectrum
        self.OldInspectResults = {} #(file, scan) => (MQScore, Annotation)  #file name only, not! path
        self.PLSDict = {} # self.PLSDict[(SpectrumFile, Scan)] = (PLS, NewPeptide)
        self.Columns = ResultsParser.Columns()
        ResultsParser.ResultsParser.__init__(self)
        
    def Main(self):
        self.CheckInputFormat(self.InputFilePath)
        if not self.InspectFormat:
            self.GetByteOffsetsForSpectra()
        MakeDirectory(self.LabeledAnnotationsDir)
        self.LabelMe()
        self.MakeOutput()


    def MakeOutput(self):
        """The results of Label.py have been put into a folder, and we now have to parse
        those and put them back into the file that people gave us.
        """
        ## get all the stuff from Label
        self.ProcessResultsFiles(self.LabeledAnnotationsDir, self.ParseLabelSpewage)
        # start putting it into the output
        Handle = open(self.InputFilePath, "rb")
        OutHandle = open(self.OutputFilePath, "wb")
        for Line in Handle.xreadlines():
            if not Line.strip():
                continue
            if Line[0] == "#":
                #header
                OutHandle.write("%s\tBetterAnnotation\tPLS\n"%Line.strip())
                continue
            Bits = Line.strip().split("\t")
            SpectrumFullPath = Bits[self.Columns.getIndex("SpectrumFile")]
            SpectrumFile = os.path.split(SpectrumFullPath)[1]
            Scan = Bits[self.Columns.getIndex("Scan#")]
            Annotation = Bits[self.Columns.getIndex("Annotation")]
            Tuple = (SpectrumFile, Scan)
            if not self.PLSDict.has_key(Tuple):
                print "NO KEY, %s, %s"%(SpectrumFile, Scan)
                continue
            (PLS, NewPeptideAnnotation) = self.PLSDict[(SpectrumFile, Scan)]
            #now write stuff out
            Bits.append("%s"%NewPeptideAnnotation)
            Bits.append("%s"%PLS)
            String = "\t".join(Bits)
            OutHandle.write("%s\n"%String)
        OutHandle.close()

    def ParseLabelSpewage(self, FilePath):
        """In each file I am going to grep out
        filename, scan number, PLS, better peptide if such exists
        """
        
        ##in the filename are the scan number and mzxml filename
        if not FilePath[-3:] == "txt":
            return #skip png images
        (Path, FileName) = os.path.split(FilePath)
        Pos = FileName.find("mzXML") + 5
        SpectrumFile = FileName[:Pos]
        Dot = FileName.find(".", Pos+1)
        Scan = FileName[Pos+1:Dot] # string value, not int
        NewPeptide = None
        Handle= open(FilePath, "rb")
        PLS = "N/A" #default, shoudl get overridden for every file
        for Line in Handle.xreadlines():
            Line = Line.strip()
            #hard coded magic
            if Line[:10] == "Phosphate ":
                #Phosphate Localization Score: 52.2
                Colon = Line.find(":")
                PLS = Line[Colon + 1:]
                #print Line
                #print "I parsed out %s"%PLS
            if Line[:7] == "WARNING":
                #parse out new peptide
                ToSplit = Line.replace("WARNING: Better annotation than input.", "")
                (BetterMQScore, NewPeptide) = ToSplit.split(",")
                NewPeptide = NewPeptide.strip()
            if Line[:2] == "b2":
                #this means we've started to get into the rest of the verbose output
                # and past what we care about
                break
        Handle.close()
        Tuple = (SpectrumFile, Scan)
        self.PLSDict[Tuple] = (PLS, NewPeptide)


    def GetByteOffsetsForSpectra(self):
        "Read mzXML from either a single file, or directory, creating the self.ScanOffset dictionary"
        Abacus = GetByteOffset.Abacus()
        if os.path.isdir(self.MZXMLDir):
            for FileName in os.listdir(self.MZXMLDir):
                (Stub, Extension) = os.path.splitext(FileName)
                if Extension.lower() == ".mzxml":
                    Path = os.path.join(self.MZXMLDir, FileName)
                    ScanOffsetSingleFile = Abacus.GetByteOffset(Path)
                    for (ScanNumber, ScanOffset) in ScanOffsetSingleFile.items():
                        self.ScanOffset[(FileName, ScanNumber)] = (Path, ScanOffset)
        else:
            ScanOffsetSingleFile = Abacus.GetByteOffset(self.MZXMLDir)
            FileName = os.path.split(self.MZXMLDir)[1]
            for (ScanNumber, ScanOffset) in ScanOffsetSingleFile.items():
                self.ScanOffset[(FileName, ScanNumber)] = (self.MZXMLDir, ScanOffset)
                #print "Storing value (%s,%s) with key (%s, %s)"%(self.MZXMLDir, ScanOffset, FileName, ScanNumber)

    def LabelMe(self):
        Handle = open(self.InputFilePath, "rb")
        Dymo = Label.LabelClass()
        Count = 0
        GoodScoreCount = 0
        WrongChargeCount = 0
        ScoredWorseCount = 0
        for Line in Handle.xreadlines():
            if Line[0] == "#":
                self.Columns.initializeHeaders(Line)
                continue
            if not Line.strip():
                continue
            Bits = list(Line.strip().split("\t"))
            #Charge = int (Bits[self.Columns.Charge])  I don't thin I need this anymore
            Count +=1
            Annotation = Bits[self.Columns.getIndex("Annotation")]
            #print "Annotation :%s:"%Annotation
            FileName = Bits[self.Columns.getIndex("SpectrumFile")]
            Scan = int(Bits[self.Columns.getIndex("Scan#")])
            if not self.InspectFormat:
                FileNameMinusPath = os.path.split(FileName)[1]
                (FullPathDummy, ByteOffset) = self.ScanOffset[(FileNameMinusPath, Scan)]
                #print (FullPathDummy, ByteOffset)
                #continue
            else:
                ByteOffset = int(Bits[self.Columns.getIndex("SpecFilePos")])
            (Path,File) = os.path.split(FileName)
            FileName = os.path.join(self.MZXMLDir, File)
            VerboseFileName = "%s.%s.%s.verbose.txt"%(File, Scan, Annotation[2:-2])
            ImageFileName = "%s.%s.%s.png"%(File, Scan, Annotation[2:-2])
            VerboseFilePath = os.path.join(self.LabeledAnnotationsDir, VerboseFileName)
            ImageFilePath = os.path.join(self.LabeledAnnotationsDir, ImageFileName)
            ## as we've got a single Dymo object, we must be passing in full args list
            ## -p to suppress the image popup, and -P for the PLS score
            Args = " -r %s -b %d -a %s -v %s -w %s -p -P"%(FileName, ByteOffset, Annotation, VerboseFilePath, ImageFilePath)
            ArgsList = Args.split()
            #print "Parsing Results for %s, scan %s, charge %s"%(FileName, Scan, Charge)
            #print "Args: %s"%Args
            Dymo.ParseCommandLine(ArgsList)
            Dymo.Main()

        Handle.close()
        
    def CheckInputFormat(self, FileName):
        """This method serves to catch input files that are not in the
        proper Inspect format.  If this is the case, then we must convert the
        files to Inspect format.  This basically means that we put a byte offset at the
        end.
        Expected format. (tab delimited, 3 columns)
        Spectrum File          Spectrum Number (int)            Annotation (string, no! numbers!)
        """
        Handle = open (self.InputFilePath, "rb")
        ## 1. get the first line and see if it's already in Inspect Format
        Line = Handle.readline()
        try:
            Bits = Line.strip().split("\t")
        except:
            print "####################################################"
            print "Input file in improper format. Please read tutorial."
            sys.exit(1)
        #if not len(Bits) < self.Columns.getIndex("SpecFilePos"):
        #    self.InspectFormat = 1
        #    return # in inspect format.  it's okay
        ## 2. Check to see if each line of the input file has the proper format
        Reject = 0
        for Line in Handle.xreadlines():
            if Line[0] == "#":
                self.Columns.initializeHeaders(Line)
                continue
            try:
                Bits = Line.strip().split("\t")
            except:
                print "####################################################"
                print "Input file in improper format. Please read tutorial."
                sys.exit(1)
            #now check to see if column 1 is a number, and 2 is a string (with no brackets)
            try:
                SpectrumNumber = int(Bits[self.Columns.getIndex("Scan#")])
            except:
                Reject = 1
                print "Second column must be a integer representing the spectrum number"
            Annotation = Bits[self.Columns.getIndex("Annotation")]
            AcceptArray = string.ascii_letters
            AcceptArray += "."  #for delimiting the prefix/suffix
            AcceptArray += "*"  # for the beginning/end of a protein. should only be in prefix/suffix
            AcceptArray += string.digits
            for Index in range(len(Annotation)):
                if not Annotation[Index] in AcceptArray:
                    print "This annotation is in an improper format %s"%Annotation
                    Reject = 1
                    break
            if Reject:
                print "####################################################"
                print "There were formatting problems with the input file"
                print "We cannot proceed.  Please read the tutorial."
                sys.exit(1)
        print "Input file %s received in the correct format"%FileName
    def ParseCommandLine(self,Arguments):
        (Options, Args) = getopt.getopt(Arguments, "r:w:m:d:")
        OptionsSeen = {}
        for (Option, Value) in Options:
            OptionsSeen[Option] = 1
            if Option == "-r":
                # -r results file(s)
                if not os.path.exists(Value):
                    print "** Error: couldn't find results file '%s'\n\n"%Value
                    print UsageInfo
                    sys.exit(1)
                self.InputFilePath = Value
            if Option == "-d":
                self.LabeledAnnotationsDir = Value
            if Option == "-w":
                self.OutputFilePath = Value
            if Option == "-m":
                self.MZXMLDir = Value
        if not OptionsSeen.has_key("-r") or not OptionsSeen.has_key("-m"):
            print UsageInfo
            sys.exit(1)


def MakeDirectory(Dir):
    if os.path.exists(Dir):
        return 
    try:
        os.makedirs(Dir)
    except:
        raise
    


if __name__ == "__main__":
    try:
        import psyco
        psyco.full()
    except:
        print "(psyco not found - running in non-optimized mode)"
    MacGyver = DetectiveClass()
    MacGyver.ParseCommandLine(sys.argv[1:])
    MacGyver.Main()



#Title:          StripPTM.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.


"""
A handy function to strip unnecessary modifications from a peptide.
"""
from Utils import *
Initialize()

AMINO_ACIDS = "ACDEFGHIKLMNOPQRSTUVWY" # O and U are included, for now.
INVALID_MASS = 99999

def StripNeedlessModifications(DB, Annotation):
    """
    Replace "*" in annotations with "-".
    Also, correct "unnecessary PTM" annotations.
    Return (DBPos, FixedAnnotation) as a tuple.
    We fix the following:
    Y.A+|Y|BCD.Z   -> YABCD
    XY.A+|XY|BCD.Z -> XYABCD
    JYZ.A+|XYZ|BCD.Z -> JYZABCD
    X.ABCD+|Y|.Y   -> ABCDY
    X.ABCD+|YZ|.YZ -> ABCDYZ
    X.ABCD+|YZJ|.YZH -> ABCDYZJ
    """
    VerboseFlag = 1
    Peptide = GetPeptideFromModdedName(Annotation)
    # Find where this peptide occurs within the database:
    Aminos = Peptide.Aminos
    if Peptide.Prefix in AMINO_ACIDS:
        Aminos = Peptide.Prefix + Aminos
    if Peptide.Suffix in AMINO_ACIDS:
        Aminos += Peptide.Suffix
    DBPos = DB.find(Aminos)
    if Peptide.Prefix in AMINO_ACIDS:
        DBPos += 1
    if not Peptide.Modifications.keys():
        # An unmodified peptide?  We don't deal with those!
        return (DBPos, Annotation.replace("*", "-"))
    # Check whether a simple endpoint-shift can abolish all
    # modifications:
    ModIndex = Peptide.Modifications.keys()[0]
    ModMass = Peptide.Modifications[ModIndex][0].Mass
    # Try a shift to the LEFT:
    if ModIndex < 3:
        FlankMass = 0
        NewAA = ""
        for ShiftCharCount in (1, 2, 3):
            if DBPos - ShiftCharCount < 0:
                break
            AA = DB[DBPos - ShiftCharCount]
            FlankMass += Global.AminoMass.get(AA, INVALID_MASS)
            NewAA = AA + NewAA # prepend to the new chars
            if abs(FlankMass - ModMass) <= 2:
                # The mass matches!  Let's shift the annotation.
                if (DBPos - ShiftCharCount > 0):
                    Prefix = DB[DBPos - ShiftCharCount - 1]
                else:
                    Prefix = "-"
                FixedAnnotation = "%s.%s%s.%s"%(Prefix, NewAA, Peptide.Aminos, Peptide.Suffix)
                if VerboseFlag:
                    print "-%d The fix is in: %s to %s"%(ShiftCharCount, Annotation, FixedAnnotation)
                return (DBPos - ShiftCharCount, FixedAnnotation.replace("*", "-"))
    # Try a shift to the RIGHT:
    if ModIndex >= len(Peptide.Aminos) - 3:
        NewAA = ""
        FlankMass = 0
        OldEndpoint = DBPos + len(Peptide.Aminos) - 1
        for ShiftCharCount in (1, 2, 3):
            if OldEndpoint + ShiftCharCount >= len(DB):
                print "Off the end of the DB with %s shifted by %d"%(Annotation, ShiftCharCount)
                continue
            AA = DB[OldEndpoint + ShiftCharCount]
            FlankMass += Global.AminoMass.get(AA, INVALID_MASS)
            NewAA += AA # append the new amino acid
            if abs(FlankMass - ModMass) <= 2:
                # The mass matches!  Let's shift the annotation.
                if (OldEndpoint + ShiftCharCount + 1)<len(DB):
                    Suffix = DB[OldEndpoint + ShiftCharCount + 1]
                else:
                    Suffix = "-"
                FixedAnnotation = "%s.%s%s.%s"%(Peptide.Prefix, Peptide.Aminos, NewAA, Suffix)
                if VerboseFlag:
                    print "+%d The fix is in: %s to %s"%(ShiftCharCount, Annotation, FixedAnnotation)
                return (DBPos, FixedAnnotation.replace("*", "-"))
    # We can't edit away the PTM.  Just fix any asterisks:
    return (DBPos, Annotation.replace("*", "-"))



# -*- coding: UTF-8 -*-
# in: ProteinDBs with and without decoys, number of samples for simulation and Inspect
# out: FDR-matrix (and several result files)

# Attention: For using Inspect, this script must be called out of the Inspect-directory! The
# "cd /home/franziska/bin/Inspect" command is called be the main routine of the script

# procedure:
# 1. sample sequences from Protein-DB (for faster simulation)
# 2. call the simulator with these sequences
# 3. convert spectra to mgf-format (input format Inspect and Biceps)
# 4. sample spectra from simulated ones
# 5. run Inspect
# 6. evaluate results from Inspect to calculate an FDR-matrix

import re
import random
from Bio import SeqIO
import os
import sys
import optparse
import time
import platform

from inspectparser import parseInspect


# 1. sampling from DB (def)
def samplProt(inpath, n):
	"""
	Sampling a given number of protein sequences from a given fasta-file

	inpath: fasta-file, protein DB which will be sampled from
	n: 		number of samples
	"""

	protIDs = [] # IDs of all proteins from db
	print 'Reading protein IDs from '+inpath+' ...'
	for record in SeqIO.parse(open(inpath, "rU"), "fasta") :
		# Add ID of this record to protIDs list
		protIDs.append(record.id)
	nprot = len(protIDs)
	print "Found %i proteins" % nprot

	if n == 0 or n > nprot:
		print "Take the complete list of %i proteins" %nprot
		return inpath
	else:
		print 'Choosing %i proteins ...' %n
		rprot = random.sample(protIDs,n)

		if inpath[-3] == '.': outpath = inpath[:-3] + "_sample" + str(n) +".fasta"
		else: outpath = inpath[:-6] + "_sample" + str(n) +".fasta"
		print 'Writing selected proteins to '+outpath+" ..."
		output_handle = open(outpath, "w")
		for record in SeqIO.parse(open(inpath, "rU"), "fasta") :
			if record.id in rprot: # for every selected protein
				SeqIO.write(record, output_handle, "fasta")
		output_handle.close()
		print "done"
		return outpath

# 2. use OpenMS simulator (def)
def run_MSSim(DBfasta, outmzML, ini="/data/NG4/anke/MSSim/mscconf_sebio02.ini", param=""):
	command = "MSSimulator -in {inp} -out {out} -ini {ini} {param}".format(inp=DBfasta, out=outmzML, ini=ini, param=param)
	print "Executing:",command
	sys.stdout.flush()
	os.system(command)
	return 1

# 3. convert mzML to mgf format (def)
def convertSpectra(mzMLfile, param=""):
	mgf = mzMLfile[:-5]+".mgf"
	command = "FileConverter -in {infile} -in_type mzML -out {outfile} -out_type mgf {param}".format(infile=mzMLfile, outfile=mgf, param=param)
	print "Executing: ",command
	sys.stdout.flush()
	os.system(command)
	return mgf

# 4. sampling from simulated spectra (def)
from readMGF import sampleMGF as samplSpecs
'''def samplSpecs(inpath, n):
	"""
	Sampling a given number of spectra from a given mgf-file

	inpath: mgf-file which will be sampled from
	n: 		number of samples
	"""

	infile = open(inpath,"r")
	specs = [] # names of all spectra

	print 'Reading specTitles ...'
	for line in infile:
		if re.search("TITLE",line):
			specs.append(line.rstrip())
	if n == 0 or n > len(specs):
		print "Take the complete list of %i spectra" %len(specs)
		return inpath
	else:
		print 'Choosing %i spectra ...' %n
		rspecs = random.sample(specs,n)
		infile.seek(0)
		outpath = inpath[:-4] + "_sample" + str(n) +".mgf"
		print 'Writing selected specs to '+outpath+" ..."
		outfile = open(outpath,"w")
		for line in infile:
			if line.rstrip() in rspecs: # for ervery selected title write spectrum to new file
				outfile.write("BEGIN IONS\n")
				outfile.write(line)
				for line2 in infile:
					if re.search("BEGIN IONS",line2): break
					outfile.write(line2)
		outfile.close()
		print "done"
		return outpath
'''

def prepDB(fastaDB, path="/home/franziska/bin/Inspect/"):
	""" 5.a prepare decoyDB input for Inspect (def)

			The function will use the decoyDB to create

			Args:
					fastaDB: decoy fasta file
					path: where inspect is put
			Returns:
					This creates two files which will be used by
					InsPecT: myDB.trie and myDB.index. Once these files are created,
					they can be reused for later InsPecT runs.
	"""
	if not os.path.exists(os.path.join(path,"PrepDB.py")):
		print "InsPecT files not found, please correct path!"
		return 0
	# PrepDB.py from inSpecT source code:
	# Convert a protein database into concatenated format.
	# Processes FASTA format.
	command = "python {path}PrepDB.py FASTA {db}".format(path=path, db=fastaDB)
	print "Executing: ",command
	sys.stdout.flush()
	os.system(command)
	return 1

# 5. match spectra against database with Inspect (def)
def run_inspect(configfile, outputfile, path="", param=""):
	current_path = os.path.dirname(os.path.realpath(__file__))
	if platform.system() == 'Windows':
			inspect_path = os.path.join(current_path, 'inspect', 'inspect.exe')
	else:
			inspect_path = os.path.join(current_path, 'inspect', 'inspect')
	if len(path) > 0:
		command = inspect_path + " -i {input} -o {output} -r {path} {param}".format(input=configfile, output=outputfile, path=path, param=param)
	else:
		command = inspect_path + " -i {input} -o {output} {param}".format(input=configfile, output=outputfile, param=param)
	print "Executing: ",command , "\n"

	sys.stdout.flush()
	os.system(command)
	return 1

# 1.-6. go all steps
def calculateSimilarityMatrix(DBs, db_path="/data/NG4/anke/proteome/", nProt=100, nSpec=1000,
	sim_out_path="/data/NG4/anke/MSSim/sampled_data/", MSSim_ini="/data/NG4/anke/MSSim/mscconf_sebio02.ini",
	inspect_config="/data/NG4/anke/Inspect/config_Inspect_MSSim.txt", inspect_dir="/home/franziska/bin/Inspect/"):
	"""
	"""
	rng = range(len(DBs))
	simMat = [ [0 for i in rng] for j in rng ] # initializing output

	configfile = inspect_config[:-4]+"_py.txt" # Inspect configuration file (final version)

	for i in rng:
		# 1. sampling from DB (run)
		prot_path = samplProt(db_path+DBs[i]+".fasta",nProt)
		# 2. use OpenMS simulator with sampled proteins (run)
		out_path = sim_out_path+DBs[i]+"_sampl"+str(nProt)+"MSSim.mzML"
		run_MSSim(prot_path, out_path, ini=MSSim_ini, param="-threads 4")
		# 3. convert mzML to mgf format
		sampl_path = convertSpectra(out_path)
		# 4. sampling from simulated spectra
		spec_path = samplSpecs(sampl_path,nSpec)
		# (*) with runInspect_user_config.runInspect_config for all DBs
		# runInspect_config(spectra=spec_path, DBs=DBs, spec_path="", db_path=db_path, inspect_dir=inspect_dir, conf=configfile, user_mods="")
		for j in rng:
			# 5. calling InSpecT (*)
			db_j = db_path+DBs[j]+"_decoy.trie"
			# 5.a create trie if necessary
			if not os.path.exists(db_j):
				prepDB(db_path+DBs[j]+"_decoy.fasta",path=inspect_dir)
			inspect_out = spec_path[:-4] +"_"+DBs[j]+"_InspectOut.txt" # (!!)
			# prepare configfile # <-- call runInspect_user_config.runInspect_config instead!! (bei 5. ansetzen, j-Schleife nur mit inspect_out behalten)
			conf_in = open(inspect_config,'r')
			conf_out = open(configfile,'w')
			for line in conf_in:
				if re.match("spectra,",line):
					conf_out.write("spectra,"+spec_path+"\n")
				elif re.match("DB,",line):
					conf_out.write("DB,"+db_j+"\n")
				else:
					conf_out.write(line)

			conf_in.close()
			conf_out.close()

			run_inspect(configfile, inspect_out, path=inspect_dir)# from 5. remove all but (!!) up to here
			# 6. evaluate results from Inspect to calculate an FDR-matrix
			simMat[i][j] = parseInspect(inspect_out, silent=False)[1]
	normSimMat = [ [0 for i in rng] for j in rng ]
	for k in rng:
		for l in rng:
			normSimMat[k][l] = simMat[k][l] / simMat[k][k]
	return simMat, normSimMat


if __name__=="__main__":
	usage = """%prog DB_LIST -d DB_DIR -p NPROT -n NSPEC -i INIFILE

Sample a given number of proteins from DB for simulation of spectra,
sample a given number of spectra from simulation for running inspect
and calculate FDR from Inspect output.

"""

	t0 = time.time()
	print time.strftime("%Y/%m/%d, %H:%M:%S: Starting whole procedure and overall time measurement", time.localtime(t0))
	sys.stdout.flush()
	# configure the parser
	optparser = optparse.OptionParser(usage=usage)
	optparser.add_option('-d', '--dbdir', type='string', dest='db_dir', default="/data/NG4/anke/proteome/", help='directory of proteinDBs. [default: %default]')
	optparser.add_option('-p', '--proteinnumber', type='int', dest='nprot', default=100, help='number of samples from proteinDB. [default: %default]')
	optparser.add_option('-n', '--numberspectra', type='int', dest='nspec', default=1000, help='number of samples from spectra. [default: %default]')
	optparser.add_option('-c', '--configfile', type='string', dest='MSSim_ini', default="/data/NG4/anke/MSSim/mscconf_sebio02.ini", help='configuration file for MSSImulator. [default: %default]')
	optparser.add_option('-i', '--inspectdir', type='string', dest='insp_dir', default="/home/franziska/bin/Inspect/", help="path to 'inspect.exe'. [default: %default]")
	optparser.add_option('-o', '--outfile', type='string', dest='out', default="/data/NG4/anke/Inspect/collectOutput.txt", help='name of file resulting matrix is written to. [default: %default]')
	# parse options and arguments
	options, args = optparser.parse_args()

	if len(args) == 1:
		db_list = args[0].split(',')
		sim_out_path = "/data/NG4/anke/MSSim/sampled_data/"
		nProt = options.nprot # nr. of sampled proteins (from DB)
		nSpec = options.nspec # nr. of sampled spectra (from simulation)
		config_path = "/data/NG4/anke/Inspect/config_Inspect_MSSim.txt" # Inspect configuration file (initial version)

		M, M_n = calculateSimilarityMatrix(DBs = db_list, db_path = options.db_dir, nProt=nProt, nSpec=nSpec, sim_out_path=sim_out_path, MSSim_ini=options.MSSim_ini, inspect_config=config_path, inspect_dir=options.insp_dir)#print db_list
	else:
		optparser.print_help()
		sys.exit(1)

	outfile = open(options.out,"a")
	outfile.write("normalized similarity matrix derived from InSpect with %i sampled spectra "%nSpec)
	outfile.write("(from simulation with %i sampled proteins): \n"%nProt)
	print "normalized: \n"
	for line in M_n:
		print line
		outfile.write(str(line)+"\n")
	outfile.write("\n\n\n")
	outfile.close()
	print time.strftime("%Y/%m/%d, %H:%M:%S: all done\n", time.localtime(time.time()))
	print "took %.3f seconds overall (that's %.3f minutes)"%(time.time() - t0, (time.time() - t0)/60)
	print "----------------------------------------------------------------------------------------------------\n\n\n"



#Title:          Learning.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.

"""
The LearnerClass is an abstract machine learner.  It can be trained,
saved, and loaded.
"""
import os
import sys
import struct
import random
import math
import traceback
import cPickle
import LDA
import RunPySVM
try:
    import PySVM
except:
    print "(Warning: PySVM not present!)"

try:
    from numpy import *
    import numpy.linalg
    FloatType = float
    MatrixMultiply = dot
    InvertMatrix = numpy.linalg.inv
except:
    print "\n* Warning: Unable to import NumPy.  Logit training not available"
    print "  Please install NumPy (see http://numpy.scipy.org/ for details)"
    print "  Error details are shown here:"
    traceback.print_exc()
    
random.seed(1)

MaxSVMFeatureCount = 500

if sys.platform == "win32":
    PATH_SVMSCALE = r"C:\libsvm\windows\svmscale.exe"
    PATH_SVMTRAIN = r"C:\libsvm\windows\svmtrain.exe"
else:
    PATH_SVMSCALE = os.path.join(os.environ["HOME"], "libsvm", "svm-scale")
    PATH_SVMTRAIN = os.path.join(os.environ["HOME"], "libsvm", "svm-train")

SQRT2PI = math.sqrt(2 * math.pi)
SQRT2 = math.sqrt(2)
Cof = [76.18009172947146, -86.50532032941677,
    24.01409824083091, -1.231739572450155, 
    0.1208650973866179e-2, -0.5395239384952e-5]

def Gamma(Z):
    X = Z
    Y = Z
    Temp = X + 5.5
    Temp -= (X + 0.5) * math.log(Temp)
    Ser = 1.000000000190015
    for J in range(6):
        Y += 1
        Ser += Cof[J] / Y
    Z = -Temp + math.log(2.5066282746310005 * Ser / X)
    return math.exp(Z)




class MixtureModelClass:
    def __init__(self, BinMultiplier = 10.0):
        self.BinMultiplier = BinMultiplier
    def Model(self, Values, Histogram = None):
        if Values:
            print "Model scores.  Range is %s...%s"%(min(Values), max(Values))
        else:
            if not Histogram.keys():
                # There's nothing to model!
                self.MinBin = 0
                self.MaxBin = 0
                self.OddsTrue = {}
                return
            print "Model scores.  Range is %s...%s"%(min(Histogram.keys()), max(Histogram.keys()))
        self.MaxCycleCount = 300
        self.VerboseFlag = 0     
        if Histogram:
            self.ScoreHistogram = Histogram
        else:
            self.ScoreHistogram = {}
            for Value in Values:
                Bin = int(round(Value * self.BinMultiplier))
                self.ScoreHistogram[Bin] = self.ScoreHistogram.get(Bin, 0) + 1
        Keys = self.ScoreHistogram.keys()
        self.MinBin = min(Keys)
        self.MaxBin = max(Keys) + 1
        self.InitializeModel()
        try:
            self.ModelDistribution()
        except:
            print "* Warning: Unable to compute p-values via mixture model"
            print "* Error trace follows:"
            traceback.print_exc()
            print "self.VarianceFalse:", self.VarianceFalse
            print "self.VarianceTrue:", self.VarianceTrue
            print "MeanFalse:", self.MeanFalse
            print "GammaOffset:", self.GammaOffset
            print "ThetaFalse:", self.ThetaFalse
            print "KFalse:", self.KFalse
    def GetOddsTrue(self, X):
##        if self.CumulativeFlag:
##            # Set our odds using the cumulative probability p(score >= X) instead of
##            # the odds that the score is in this bin.
##            ErfArg = (X - self.MeanTrue) / (self.StdDevTrue * SQRT2)
##            NormalCDF = 0.5 + 0.5 * PyInspect.erf(ErfArg)
##            GX = max(0.01, X + self.GammaOffset)
##            GammaCDF = PyInspect.GammaIncomplete(self.KFalse, GX / self.ThetaFalse) #/ Gamma(self.KFalse)
##            TrueNormal = 1.0 - NormalCDF
##            FalseGamma = 1.0 - GammaCDF
##        else:
        Pow = - ((X - self.MeanTrue)**2) / (2 * self.VarianceTrue)
        TrueNormal = math.exp(Pow) / (self.StdDevTrue * SQRT2PI)
        GX = max(0.01, X + self.GammaOffset)
        
        
        FalseGamma = math.pow(GX, self.KFalse - 1) * math.exp(-GX / self.ThetaFalse) / self.GammaDemonFalse
        # Special patch-up code:
        # Toward the edges of the mixture model, odd behavior may occur where one curve falls off
        # slower than the other.  We force very low scores to get a bad odds-true, and very
        # high scores to get a good odds-true.
        if X < self.MeanTrue - self.VarianceTrue:
            FalseGamma = max(FalseGamma, 0.001)
        if X > self.MeanTrue + self.VarianceTrue:
            TrueNormal = max(TrueNormal, 0.001)
        OddsTrue = (TrueNormal * self.PriorProbabilityTrue) / (TrueNormal * self.PriorProbabilityTrue + FalseGamma * (1 - self.PriorProbabilityTrue))
        return OddsTrue
    def InitializeModel(self):
        # Initialize mixture model:
        MinValue = self.MinBin / self.BinMultiplier
        MaxValue = self.MaxBin / self.BinMultiplier
        self.MeanFalse = MinValue + (MaxValue - MinValue) * 0.25
        self.MeanTrue = MaxValue - (MaxValue - MinValue) * 0.25
        self.VarianceFalse = (MaxValue - MinValue) * 0.1
        self.VarianceTrue = (MaxValue - MinValue) * 0.1
        self.PriorProbabilityTrue = 0.1
        if MinValue < 0:
            self.GammaOffset = -MinValue
        elif MinValue > 0.1:
            self.GammaOffset = -MinValue
        else:
            self.GammaOffset = 0
        self.OddsTrue = {}
    def ModelDistribution(self):
        self.ThetaFalse = self.VarianceFalse / (self.MeanFalse + self.GammaOffset)
        self.StdDevTrue = math.sqrt(self.VarianceTrue)
        self.KFalse = (self.MeanFalse + self.GammaOffset) / self.ThetaFalse
        self.GammaDemonFalse = math.pow(self.ThetaFalse, self.KFalse) * Gamma(self.KFalse)
        for Cycle in range(self.MaxCycleCount):
            self.Cycle = Cycle
            self.EstimateOddsTrue()
            self.ComputeDistributionParameters()
    def EstimateOddsTrue(self):
        """
        One half of the E/M cycle: Estimate the probability true for each bin.
        """
        # For each bin, compute the probability that it's true:
        BestOddsTrue = 0
        for Bin in range(self.MinBin, self.MaxBin):
            X = Bin / self.BinMultiplier
            self.OddsTrue[Bin] = self.GetOddsTrue(X)
            # Somewhat hacky: If the left tail of the normal distribution falls off more slowly
            # than that of the gamma distribution, the value of OddsTrue is often rather
            # high for these (very bad!) bins.  We fix that.
            if Bin < 0:
                self.OddsTrue[Bin] = min(self.OddsTrue[Bin], 1 / float(-Bin))
            # Somewhat hacky: If the right tail of the normal distribution falls off too quickly,
            # then the odds true will decay:
            BestOddsTrue = max(BestOddsTrue, self.OddsTrue[Bin])
            if X >= self.MeanTrue:
                self.OddsTrue[Bin] = max(BestOddsTrue, self.OddsTrue[Bin])
            #print "%s: %s"%(X, self.OddsTrue[Bin])
    def ComputeDistributionParameters(self):
        """
        One half of the E/M cycle: Optimize the distribution parameters.
        """
        # Compute the new mean and variance for the true and the false distributions:
        self.CountTrue = 0
        self.MeanTrue = 0
        self.CountFalse = 0
        self.MeanFalse = 0
        for Bin in range(self.MinBin, self.MaxBin):
            X = Bin / self.BinMultiplier
            Count = self.ScoreHistogram.get(Bin, 0)
            self.MeanTrue += X * self.OddsTrue[Bin] * Count
            self.CountTrue += self.OddsTrue[Bin] * Count
            self.MeanFalse += X * (1.0 - self.OddsTrue[Bin]) * Count
            self.CountFalse += (1.0 - self.OddsTrue[Bin]) * Count
        if self.CountTrue <= 0 or self.CountFalse <= 0:
            print "** Error: Unable to fit mixture model.  Appears to be %s true and %s false matches."%(self.CountTrue, self.CountFalse)
            return 0
        self.MeanTrue /= self.CountTrue
        self.MeanFalse /= self.CountFalse
        self.PriorProbabilityTrue = self.CountTrue / (self.CountTrue + self.CountFalse)
        # Adjust GammaOffset, if the false distribution's mean is getting close to 0:
        if self.MeanFalse + self.GammaOffset < 0.1:
            print "False distribution mean is small; BUMP gamma offset up"
            self.GammaOffset += 0.5
        ##################################
        # Compute the new variation for the true and the false distributions:
        self.VarianceTrue = 0
        self.VarianceFalse = 0
        for Bin in range(self.MinBin, self.MaxBin):
            X = Bin / self.BinMultiplier
            Count = self.ScoreHistogram.get(Bin, 0)
            try:
                self.VarianceTrue += (X - self.MeanTrue)**2 * Count * self.OddsTrue[Bin]
                self.VarianceFalse += (X - self.MeanFalse)**2 * Count * (1.0 - self.OddsTrue[Bin])
            except:
                print X
                print self.MeanTrue
                print self.MeanFalse
                print self.OddsTrue[Bin]
                raise
        self.VarianceTrue /= self.CountTrue
        self.StdDevTrue = math.sqrt(self.VarianceTrue)
        self.VarianceFalse /= self.CountFalse
        #print "  True mean %.4f var %.4f"%(self.MeanTrue, self.VarianceTrue)
        #print " False mean %.4f var %.4f"%(self.MeanFalse, self.VarianceFalse)
        self.ThetaFalse = self.VarianceFalse / (self.MeanFalse + self.GammaOffset)
        self.KFalse = (self.MeanFalse + self.GammaOffset) / self.ThetaFalse
        self.GammaDemonFalse = math.pow(self.ThetaFalse, self.KFalse) * Gamma(self.KFalse)
        if self.VerboseFlag:
            print "-----------------------"
            print "Cycle %s report:"%self.Cycle
            print "Theta %.4f K %.4f GammaDenominator %.8f GammaOffset %.2f"%(self.ThetaFalse, self.KFalse, self.GammaDemonFalse, self.GammaOffset)
            print "True: Count %s mean %s variance %s"%(self.CountTrue, self.MeanTrue, self.VarianceTrue)
            print "False: Count %s mean %s variance %s"%(self.CountFalse, self.MeanFalse, self.VarianceFalse)
            print "Prior probability true: %s"%self.PriorProbabilityTrue
    def PlotDistribution(self, FileName):
        File = open(FileName, "wb")
        Header = "Bin\tValue\tHistogram\tOddsTrue\tTrueNormal\tFalseGamma\tMixture\t"
        File.write(Header + "\n")
        for Bin in range(self.MinBin, self.MaxBin):
            Str = "%s\t%s\t"%(Bin, Bin / self.BinMultiplier)
            Str += "%s\t%s\t"%(self.ScoreHistogram.get(Bin, 0), self.OddsTrue[Bin])
            X = Bin / self.BinMultiplier
            # Plot gamma and normal curves:
            Pow = - ((X - self.MeanTrue)**2) / (2 * self.VarianceTrue)
            TrueNormal = math.exp(Pow) / (self.StdDevTrue * SQRT2PI)
            GX = max(0.01, X + self.GammaOffset)
            FalseGamma = math.pow(GX, self.KFalse - 1) * math.exp(-GX / self.ThetaFalse) / self.GammaDemonFalse
            Str += "%s\t%s\t"%(TrueNormal, FalseGamma)
##            # Plot gamma and noraml CDF:
##            ErfArg = (X - self.MeanTrue) / (self.StdDevTrue * SQRT2)
##            NormalCDF = 0.5 + 0.5 * PyInspect.erf(ErfArg)
##            GX = max(0.01, X + self.GammaOffset)
##            GammaCDF = PyInspect.GammaIncomplete(self.KFalse, GX / self.ThetaFalse) #/ Gamma(self.KFalse)
##            Str += "%s\t%s\t"%(NormalCDF, GammaCDF)
            MergedMixture = TrueNormal * self.PriorProbabilityTrue
            MergedMixture += FalseGamma * (1.0 - self.PriorProbabilityTrue)
            Str += "%s\t"%(MergedMixture)
            File.write(Str + "\n")
    def PickleSelf(self, File):
        cPickle.dump(self.BinMultiplier, File)
        cPickle.dump(self.PriorProbabilityTrue, File)
        cPickle.dump(self.MeanTrue, File)
        cPickle.dump(self.VarianceTrue, File)
        cPickle.dump(self.StdDevTrue, File)
        cPickle.dump(self.GammaOffset, File)
        cPickle.dump(self.KFalse, File)
        cPickle.dump(self.ThetaFalse, File)
        cPickle.dump(self.GammaDemonFalse, File)

def UnpickleMixtureModel(File):
    Model = MixtureModelClass()
    Model.BinMultiplier = cPickle.load(File)
    Model.PriorProbabilityTrue = cPickle.load(File)
    Model.MeanTrue = cPickle.load(File)
    Model.VarianceTrue = cPickle.load(File)
    Model.StdDevTrue = cPickle.load(File)
    Model.GammaOffset = cPickle.load(File)
    Model.KFalse = cPickle.load(File)
    Model.ThetaFalse = cPickle.load(File)
    Model.GammaDemonFalse = cPickle.load(File)
    return Model

class FeatureVector:
    def __init__(self):
        self.FileBits = []
        self.Features = []
        self.ScaledFeatures = None
        self.TrueFlag = 0
        self.Score = 0 # as assigned by an owning model
        
class FeatureSetClass:
    """
    A feature-set is a list of TRUE tuples, and a list of FALSE tuples.  Normally there
    is a FeatureSetClass for testing, and one for training.
    """
    def __init__(self):
        self.TrueVectors = []
        self.FalseVectors = []
        self.AllVectors = []
        self.TrueCount = 0
        self.FalseCount = 0
        self.Count = 0
        self.PriorProbabilityFalse = 0.5
    def SetCounts(self):
        self.Count = len(self.AllVectors)
        self.TrueCount = len(self.TrueVectors)
        self.FalseCount = len(self.FalseVectors)
        if len(self.AllVectors):
            self.Size = len(self.AllVectors[0].Features)
    def FindFeatureRanges(self):
        """
        Simple scaling function: Find min and max values to push features into [-1, 1]
        """
        Values = []
        Vector = self.AllVectors[0]
        Size = len(Vector.Features)
        print "SIZE:", Size
        for X in range(Size):
            Values.append([])
        for Vector in self.AllVectors:
            for X in range(Size):
                Values[X].append(Vector.Features[X])
        self.MinValues = []
        self.MaxValues = []
        for X in range(Size):
            Values[X].sort()
            ValIndex = int(round(len(Values[X]) * 0.025))
            MinValue = Values[X][ValIndex]
            self.MinValues.append(MinValue)
            ValIndex = int(round(len(Values[X]) * 0.975))
            MaxValue = Values[X][ValIndex]
            self.MaxValues.append(MaxValue)
##        print "Range:"
##        for X in range(Size):
##            print "%s: %.4f-%.4f"%(X, self.MinValues[X], self.MaxValues[X])
        pass
    def ScaleFeatures(self):
        """
        Simple scaling function: Pushes 90% of feature values into the range [-1, 1].
        Assumes that self.MinValues and self.MaxValues are set!
        """
        self.Size = len(self.AllVectors[0].Features)
        for Vector in self.AllVectors:
            Vector.ScaledFeatures = [0]*self.Size
        for X in range(self.Size):
            HalfRange = (self.MaxValues[X] - self.MinValues[X]) / 2.0
            if not HalfRange:
                continue
            #print "Feature %s: Range %s...%s"%(X, self.MinValues[X], self.MaxValues[X])
            for Vector in self.AllVectors:
                Vector.ScaledFeatures[X] = (Vector.Features[X] - self.MinValues[X]) / HalfRange - 1.0
    def __str__(self):
        return "<%sT %sF>"%(len(self.TrueVectors), len(self.FalseVectors))
    def GetPriorProbabilityFalse(self, DBTrueToFalseRatio):
        """
        Compute the prior probability that an arbitrary peptide is false.
        """
        # In a 1:1 database, there's 1 bogus peptide in a valid protein
        # for each bogus peptide in an invalid protein; in that case, DBTrueToFalseRatio is 1.0
        # In a 1:99 database, the ratio is 1/99.
        FalseWithinTrue = self.FalseCount * DBTrueToFalseRatio
        if FalseWithinTrue >= self.TrueCount:
            # Uh-oh; there are FEWER peptides from valid proteins than we would expect
            # to see by chance!  Let's (arbitrarily) cap the prior probability false
            # at 99%.
            print "Warning: FalseWithinTrue = %s >= %s!"%(FalseWithinTrue, self.TrueCount)
            self.PriorProbabilityFalse = 0.99
            return
        VectorCount = len(self.AllVectors)
        self.PriorProbabilityFalse = (VectorCount - (self.TrueCount - FalseWithinTrue)) / float(VectorCount)
        print "==>>PriorProbabilityFalse: %s"%(self.PriorProbabilityFalse)
    def SaveTabDelimited(self, File):
        if type(File) == type(""):
            File = open(File, "wb")
            CloseFlag = 1
        else:
            CloseFlag = 0
        #File = open(FileName, "wb")
        for VectorIndex in range(len(self.AllVectors)):
            Vector = self.AllVectors[VectorIndex]
            String = "%s\t%s\t"%(VectorIndex, Vector.TrueFlag)
            for Value in Vector.Features:
                String += "%s\t"%Value
            File.write(String + "\n")
        if CloseFlag:
            File.close()

def LoadGeneralModel(FileName):
    File = open(FileName, "rb")
    ModelType = cPickle.load(File)
    File.close()
    if ModelType == "LDA":
        Model = LDAModel()
        Model.LoadModel(FileName)
    elif ModelType == "SVM":
        Model = SVMModel()
        Model.LoadModel(FileName)
    elif ModelType == "LOGIT":
        Model = LogitModel()
        Model.LoadModel(FileName)
    else:
        print "** Error: Unable to load model type '%s'"%ModelType
        return None
    return Model

class LearnerClass:
    def __init__(self, FeatureList = None):
        # The entries in FeatureList are indices into the
        # available features of our training and testing sets.
        self.FeatureList = FeatureList
        # OddsTrue[Bin] = probability that an instance with a score
        # in this bin or HIGHER is correct.
        self.OddsTrue = {}
        self.PValue = {}
        # Bin = int(round(Score * self.BinScalingFactor))
        self.BinScalingFactor = 10
        self.MixtureModel = None
    def SaveModel(self, FileName):
        raise ValueError, "Abstract method - override in subclass!"
    def LoadModel(self, FileName):
        raise ValueError, "Abstract method - override in subclass!"
    def Train(self, FeatureSet):
        raise ValueError, "Abstract method - override in subclass!"
    def Test(self, FeatureSet):
        raise ValueError, "Abstract method - override in subclass!"
    def ReportROC(self, FeatureSet, OutputFileName = None):
        SortedList = []
        for Vector in FeatureSet.AllVectors:
            SortedList.append((Vector.Score, random.random(), Vector))
        SortedList.sort()
        SortedList.reverse()
        OverallTrueCount = 0
        OverallFalseCount = 0
        # If there are many many vectors, then we'll end up with an unwieldy curve that's
        # awkward to plot.  So, consider PHASING things:
        Slice = (len(SortedList) / 30000) + 1 # 2 or larger
        OldSortedList = SortedList
        SortedList = []
        print "SLICE roc-curve list: Take every %sth entry"%Slice
        for X in range(len(OldSortedList)):
            if X % Slice == 0:
                SortedList.append(OldSortedList[X])
                Vector = OldSortedList[X][-1]
                if Vector.TrueFlag:
                    OverallTrueCount += 1
                else:
                    OverallFalseCount += 1
        OldSortedList = None
        TrueCount = 0
        FalseCount = 0
        if OutputFileName:
            ROCCurvePlotFile = open(OutputFileName, "wb")
        RowCount = 0
        ROCTPForFP = {}
        ROCTPForFPCount = {}
        for (Score, Dummy, Vector) in SortedList:
            RowCount += 1
            if Vector.TrueFlag:
                TrueCount += 1
            else:
                FalseCount += 1
            OverallTPRate = TrueCount / float(max(1, OverallTrueCount))
            OverallFPRate = FalseCount / float(max(1, OverallFalseCount))
            Bin = int(round(OverallFPRate * 100))
            ROCTPForFP[Bin] = ROCTPForFP.get(Bin, 0) + OverallTPRate
            ROCTPForFPCount[Bin] = ROCTPForFPCount.get(Bin, 0) + 1
            if OutputFileName:
                ROCCurvePlotFile.write("%s\t%s\t%s\t%s\t%s\t\n"%(RowCount, TrueCount, FalseCount, OverallFPRate, OverallTPRate))
        if OutputFileName:
            ROCCurvePlotFile.close()
        # Compute the area under the ROC curve.
        for Bin in range(0, 100):
            if ROCTPForFP.has_key(Bin):
                ROCTPForFP[Bin] /= float(ROCTPForFPCount[Bin])
        ROCArea = 0
        for Bin in range(0, 100):
            if ROCTPForFP.has_key(Bin):
                ROCArea += 0.01 * ROCTPForFP[Bin]
                #print "%s: %s"%(Bin, ROCTPForFP[Bin])
            else:
                # Interpolate between points:
                PrevX = 0 # default
                PrevY = 0 # default
                for PrevBin in range(Bin - 1, -1, -1):
                    if ROCTPForFP.has_key(PrevBin):
                        PrevX = PrevBin
                        PrevY = ROCTPForFP[PrevBin]
                        break
                NextX = 100
                NextY = 1
                for NextBin in range(Bin + 1, 101):
                    if ROCTPForFP.has_key(NextBin):
                        NextX = NextBin
                        NextY = ROCTPForFP[NextBin]
                        break
                InterpolatedValue = PrevY + (Bin - PrevX) * float(NextY - PrevY) / (NextX - PrevX)
                ROCArea += 0.01 * InterpolatedValue
        print "ROC curve area:", ROCArea
    def ReportAccuracy(self, FeatureSet, ROCFilePath = None):
        """
        Called after Test(FeatureSet), to measure how well we did at separating
        the true and false vectors by score.  Compute OddsTrue, as well.
        """
        SortedList = []
        for Vector in FeatureSet.AllVectors:
            SortedList.append((Vector.Score, Vector))
        # sort from HIGHEST to LOWEST score:
        SortedList.sort()
        SortedList.reverse()
        #self.ComputeOddsTrue(SortedList)
        self.ComputePValues(SortedList)
        #self.ComputePValues(SortedList, FeatureSet.PriorProbabilityFalse)
        Rates = [0.05, 0.01, 0.1, 0.5, 0.005]
        CountsByRate = [0, 0, 0, 0, 0, 0]
        CumulativeTrue = 0
        CumulativeFalse = 0
        Cumulative = 0
        PrevScore = None
        for (Score, Vector) in SortedList:
            if Vector.TrueFlag:
                CumulativeTrue += 1
            else:
                CumulativeFalse += 1
            Cumulative += 1
            FractionFalse = CumulativeFalse / float(Cumulative)
            if Score != PrevScore:
                for RateIndex in range(len(Rates)):
                    if FractionFalse < Rates[RateIndex]:
                        CountsByRate[RateIndex] = CumulativeTrue
                #print Score, FractionFalse, CumulativeTrue, CumulativeFalse
            PrevScore = Score
        print "Counts by FDRate: %d@5%% %d@1%% %d@10%% %d@50%%"%(CountsByRate[0], CountsByRate[1], CountsByRate[2], CountsByRate[3])
        SensitivityByRate = []
        for Count in CountsByRate:
            SensitivityByRate.append(100 * Count / float(max(1, FeatureSet.TrueCount)))
        print "FeatureSet.TrueCount:", FeatureSet.TrueCount
        print "  Sensitivity: %.2f@5%% %.2f@1%% %.2f@10%% %.2f@50%%"%(SensitivityByRate[0], SensitivityByRate[1], SensitivityByRate[2], SensitivityByRate[3])
        if ROCFilePath:
            self.ReportROC(FeatureSet, ROCFilePath)
        return CountsByRate
    def ComputeOddsTrue(self, SortedList):
        "DEPRECATED; use ComputePValue instead"
        if len(SortedList) < 200:
            BlockSize = len(SortedList) / 4
        else:
            BlockSize = 100
        WindowTrueSum = 0
        WindowFalseSum = 0
        for Entry in SortedList[:BlockSize - 1]:
            if Entry[1].TrueFlag:
                WindowTrueSum += 1
            else:
                WindowFalseSum += 1
        for Index in range(len(SortedList)):
            # Add one entry to the window:
            if Index + BlockSize < len(SortedList):
                Entry = SortedList[Index + BlockSize]
                if Entry[1].TrueFlag:
                    WindowTrueSum += 1
                else:
                    WindowFalseSum += 1
            # Compute the probability-true for this window:
            Vector = SortedList[Index][1]
            OddsTrue = WindowTrueSum / float(WindowTrueSum + WindowFalseSum)
            Bin = int(round(Vector.Score * self.BinScalingFactor))
            self.OddsTrue[Bin] = OddsTrue
            # Remove leftmost entry from the window:
            if Index >= BlockSize:
                Entry = SortedList[Index - BlockSize]
                if Entry[1].TrueFlag:
                    WindowTrueSum -= 1
                else:
                    WindowFalseSum -= 1
    def GetPValue(self, Score):
        if self.MixtureModel:
            return 1.0 - self.MixtureModel.GetOddsTrue(Score)
        Bin = int(round(Score * self.BinScalingFactor))
        Keys = self.PValue.keys()
        MinKey = min(Keys)
        MaxKey = max(Keys)
        if Bin < MinKey:
            return self.PValue[MinKey]
        if Bin > MaxKey:
            return self.PValue[MaxKey]
        return self.PValue[Bin]
    def ComputePValuesMixtureModel(self, SortedList):
        """
        Our feature-set has an empirical distribution of scores.  We'll approximate
        this distribution as a mixture of two distributions: gamma (false) and
        normal (true).  Then we'll derive p-value (probability false) for each
        score-bin.
        """
        Scores = []

        
        for (Score, Vector) in SortedList:
            Scores.append(Score)
        #for Bin in range(Model.MinBin, Model.MaxBin):
        #    self.PValue[Bin] = 1.0 - Model.OddsTrue[Bin]
        self.MixtureModel = MixtureModelClass(self.GetMMBinMultiplier())
        self.MixtureModel.Model(Scores)
        for (Score, Vector) in SortedList:
            Vector.PValue = 1.0 - self.MixtureModel.GetOddsTrue(Score)        
        self.MixtureModel.PlotDistribution("PValues.txt")
    def ComputePValues(self, SortedList):
        self.ComputePValuesMixtureModel(SortedList)
    def ComputePValuesEmpirical(self, SortedList, PriorProbabilityFalse):
        """
        DEPRECATED - called only if mixture model fails.
        
        Input SortedList is a list of the form (ModelScore, FeatureVector), sorted from
        highest to lowest ModelScore.
        
        self.PValue[Bin] is the probability that a peptide P is FALSE, given a score of
        Bin or better.  Formally, it equals P(not P | S(P)>=Bin).  By Bayes' Theorem, this
        is equal to:
        P(S(P)>=Bin | not P) * P(not P) / P(S(P)>=Bin) 
        """
        # Passed in: PriorProbabilityFalse == P(not P)
        # And set up dictionaries PValueTotals / PValueCounts to
        #   compute ProbRatio == P(S(P)>=Bin | not P) / P(S(P)>=Bin)
        #PriorProbabilityFalse = 0
        TrueInstanceCount = 0
        HighScoringCount = 0
        HighScoringFalseInstanceCount = 0
        TotalInstances = len(SortedList)
        PValueTotals = {}
        PValueCounts = {}
        TotalFalseInstances = 0
        for (Score, Vector) in SortedList:
            if not Vector.TrueFlag:
                TotalFalseInstances += 1
        HistoGood = {}
        HistoBad = {}
        TempOutputFile = open("TempPTMPValue.txt", "wb")
        TempOutputFile.write("Bin\tHighFalse\tTotalInstances\tTotalFalseInstances\tHighScoringCount\t\n")
        for (Score, Vector) in SortedList:
            HighScoringCount += 1
            if Vector.TrueFlag:
                TrueInstanceCount += 1
            else:
                HighScoringFalseInstanceCount += 1
            ProbRatio = (HighScoringFalseInstanceCount * TotalInstances) / float(TotalFalseInstances * HighScoringCount)
            #ProbRatio = FalseInstanceCount / float(TrueInstanceCount + FalseInstanceCount)
            Bin = int(round(Score * self.BinScalingFactor))
            PValueTotals[Bin] = PValueTotals.get(Bin, 0) + ProbRatio
            PValueCounts[Bin] = PValueCounts.get(Bin, 0) + 1
            TempOutputFile.write("%s\t%s\t%s\t%s\t%s\t\n"%(Bin, HighScoringFalseInstanceCount, TotalInstances, TotalFalseInstances, HighScoringCount))
            if Vector.TrueFlag:
                HistoGood[Bin] = HistoGood.get(Bin, 0) + 1
            else:
                HistoBad[Bin] = HistoBad.get(Bin, 0) + 1
        #PriorProbabilityFalse = FalseInstanceCount / float(FalseInstanceCount + TrueInstanceCount)
        Keys = PValueTotals.keys()
        Keys.sort()
        for Bin in Keys:
            AverageProbRatio = PValueTotals[Bin] / float(PValueCounts[Bin])
            self.PValue[Bin] = AverageProbRatio * PriorProbabilityFalse
            self.PValue[Bin] = max(self.PValue[Bin], 0.0001)
            print "%s: %s"%(Bin, self.PValue[Bin])
        TempOutputFile.write("\n\n\n")
        for Bin in Keys:
            TempOutputFile.write("%s\t%s\t%s\t\n"%(Bin, HistoGood.get(Bin, 0), HistoBad.get(Bin, 0)))
        ############################################################
        # Interpolate p-values, for missing bins:
        Keys = self.PValue.keys()
        Keys.sort()
        MinKey = min(Keys)
        MaxKey = max(Keys)
        for Bin in range(MinKey, MaxKey):
            if self.PValue.has_key(Bin):
                PrevBin = Bin
                PrevPValue = self.PValue[Bin]
                continue
            # Find the next bin:
            for NextBin in range(Bin + 1, MaxKey + 1):
                if self.PValue.has_key(NextBin):
                    NextPValue = self.PValue[NextBin]
                    break
            # Interpolate from (PrevBin, PrevPValue) to (NextBin, NextPValue):
            Slope = (NextPValue - PrevPValue) / float(NextBin - PrevBin)
            Intermediate = PrevPValue + Slope * (Bin - PrevBin)
            self.PValue[Bin] = Intermediate
    def GetMMBinMultiplier(self):
        return 10.0 #default


class LDAModel(LearnerClass):
    def __init__(self, FeatureList = None):
        if FeatureList:
            self.Size = len(FeatureList)
        LearnerClass.__init__(self, FeatureList)
    def GetCovarianceArray(self, VectorList):
        VectorCount = float(len(VectorList))
        C = zeros((self.Size, self.Size), FloatType)
        for Vector in VectorList:
            for X in range(self.Size):
                for Y in range(self.Size):
                    C[X][Y] += Vector[X] * Vector[Y] / VectorCount
        return C
    def LoadModel(self, FileName):
        File = open(FileName, "rb")
        cPickle.load(File) # model type
        self.FeatureList = cPickle.load(File)
        self.PValue = cPickle.load(File)
        self.MinValues = cPickle.load(File)
        self.MaxValues = cPickle.load(File)
        self.CI = cPickle.load(File)
        self.MeanGood = cPickle.load(File)
        self.ConstantGood = cPickle.load(File)
        self.MeanBad = cPickle.load(File)
        self.ConstantBad = cPickle.load(File)
        self.Size = len(self.FeatureList)
        self.MixtureModel = UnpickleMixtureModel(File)
        # Verbose stuff:
        print "\n>>>PyLoadLDAModel(%s)"%FileName
        print "Features: %s"%self.Size
        print "MinValues: %.4f...%.4f"%(self.MinValues[0], self.MinValues[-1])
        print "MaxValues: %.4f...%.4f"%(self.MaxValues[0], self.MaxValues[-1])
        print "MeanGood: %.4f...%.4f"%(self.MeanGood[0], self.MeanGood[-1])
        print "MeanBad: %.4f...%.4f"%(self.MeanBad[0], self.MeanBad[-1])
        if self.Size > 1:
            print "CI: %.4f, %.4f...%.4f, %.4f"%(self.CI[0][0], self.CI[0][1],
                self.CI[self.Size - 1][self.Size - 2],
                self.CI[self.Size - 1][self.Size - 1]
                                         )
        print "ConstTrue %.4f, ConstFalse %.4f"%(self.ConstantGood, self.ConstantBad)
        File.close()
    def SaveBinaryModel(self, FileName):
        """
        Write out a binary representation of this model.  
        """
        File = open(FileName, "wb")
        File.write(struct.pack("<i", self.Size))
        for FeatureIndex in range(self.Size):
            File.write(struct.pack("<d", self.MinValues[FeatureIndex]))
        for FeatureIndex in range(self.Size):
            File.write(struct.pack("<d", self.MaxValues[FeatureIndex]))
        for FeatureIndex in range(self.Size):
            File.write(struct.pack("<d", self.MeanGood[FeatureIndex]))
        for FeatureIndex in range(self.Size):
            File.write(struct.pack("<d", self.MeanBad[FeatureIndex]))
        File.write(struct.pack("<d", self.ConstantGood))
        File.write(struct.pack("<d", self.ConstantBad))
        for Row in range(self.Size):
            for Column in range(self.Size):
                File.write(struct.pack("<d", self.CI[Row][Column]))
        File.close()
    def SaveModel(self, FileName):
        File = open(FileName, "wb")
        cPickle.dump("LDA", File)
        cPickle.dump(self.FeatureList, File)
        cPickle.dump(self.PValue, File)
        cPickle.dump(self.MinValues, File)
        cPickle.dump(self.MaxValues, File)
        cPickle.dump(self.CI, File)
        cPickle.dump(self.MeanGood, File)
        cPickle.dump(self.ConstantGood, File)
        cPickle.dump(self.MeanBad, File)
        cPickle.dump(self.ConstantBad, File)
        try:
            self.MixtureModel.PickleSelf(File)
        except:
            cPickle.dump(None, File)
        File.close()
    def ScoreInstance(self, RawFeatures):
        Features = []
        for FeatureIndex in self.FeatureList:
            Features.append(RawFeatures[FeatureIndex])
        for FeatureIndex in range(self.Size):
            X = self.FeatureList[FeatureIndex]
            HalfRange = (self.MaxValues[X] - self.MinValues[X]) / 2.0
            if not HalfRange:
                continue
            Features[FeatureIndex] = (Features[FeatureIndex] - self.MinValues[X]) / HalfRange - 1.0
        CIProduct = MatrixMultiply(self.CI, Features)
        ReadingGood = MatrixMultiply(self.MeanGood, CIProduct) + self.ConstantGood
        ReadingBad = MatrixMultiply(self.MeanBad, CIProduct) + self.ConstantBad
        return (ReadingGood - ReadingBad)
    def Test(self, FeatureSet):
        # Scale features, according to our trained scaling
        FeatureSet.MinValues = self.MinValues
        FeatureSet.MaxValues = self.MaxValues
        FeatureSet.ScaleFeatures()
        # Compute scores:
        for Vector in FeatureSet.AllVectors:
            FixedVector = []
            for FeatureIndex in self.FeatureList:
                FixedVector.append(Vector.ScaledFeatures[FeatureIndex])
            CIProduct = MatrixMultiply(self.CI, FixedVector)
            ReadingGood = MatrixMultiply(self.MeanGood, CIProduct) + self.ConstantGood
            ReadingBad = MatrixMultiply(self.MeanBad, CIProduct) + self.ConstantBad
            Vector.Score = (ReadingGood - ReadingBad)
    def Train(self, FeatureSet, VerboseFlag = 0):
        # Get the feature range (training only):
        FeatureSet.FindFeatureRanges()
        self.MinValues = FeatureSet.MinValues
        self.MaxValues = FeatureSet.MaxValues
        # Sanity-checking: If a feature's range is a single point,
        # then it's not useful - AND, it will generate a non-invertible
        # matrix.  So, let's filter out any such features.
        InputFeatureList = self.FeatureList
        self.FeatureList = []
        for FeatureIndex in InputFeatureList:
            if self.MinValues[FeatureIndex] < self.MaxValues[FeatureIndex]:
                self.FeatureList.append(FeatureIndex)
            else:
                print "* Warning: Discarding feature '%s', every entry is %s"%(FeatureIndex, self.MinValues[FeatureIndex])
        self.Size = len(self.FeatureList)
        if self.Size == 0:
            print "<< no features - bailing out >>"
            return
        # Scale features:
        FeatureSet.ScaleFeatures()
        AllVectors = []
        TrueVectors = []
        FalseVectors = []
        for Vector in FeatureSet.AllVectors:
            ScaledVector = []
            for FeatureIndex in self.FeatureList:
                ScaledVector.append(Vector.ScaledFeatures[FeatureIndex])
            AllVectors.append(ScaledVector)
            if Vector.TrueFlag:
                TrueVectors.append(ScaledVector)
            else:
                FalseVectors.append(ScaledVector)
        print "First true vector:", TrueVectors[0]
        print "First false vector:", FalseVectors[0]
##        # Temp: Ensure the vector lists are the same size!
##        VectorCount = min(len(TrueVectors), len(FalseVectors))
##        random.shuffle(FalseVectors)
##        FalseVectors = FalseVectors[:VectorCount]
        ############################################################
        # Compute the mean vectors (training only):
        self.MeanGood = [0] * self.Size
        self.MeanBad = [0] * self.Size
        self.MeanGlobal = [0] * self.Size
        for Vector in TrueVectors:
            for X in range(self.Size):
                self.MeanGlobal[X] += Vector[X] / float(FeatureSet.Count)
                self.MeanGood[X] += Vector[X] / float(FeatureSet.TrueCount)
        for Vector in FalseVectors:
            for X in range(self.Size):
                self.MeanGlobal[X] += Vector[X] / float(FeatureSet.Count)
                self.MeanBad[X] += Vector[X] / float(FeatureSet.FalseCount)
        if VerboseFlag:
            print "MeanGood:\n  ",
            for Value in self.MeanGood:
                print "%.3f"%Value,
            print
            print "MeanBad:\n  ",
            for Value in self.MeanBad:
                print "%.3f"%Value,
            print
            print "MeanGlobal:\n  ",
            for Value in self.MeanGlobal:
                print "%.3f"%Value,
            print
        ############################################################
        # Compute mean-corrected vectors:
        MeanCorrectedGoodVectors = []
        MeanCorrectedBadVectors = []
        for Vector in TrueVectors:
            NewVector = []
            for X in range(self.Size):
                NewVector.append(Vector[X] - self.MeanGlobal[X])
            MeanCorrectedGoodVectors.append(NewVector)
        for Vector in FalseVectors:
            NewVector = []
            for X in range(self.Size):
                NewVector.append(Vector[X] - self.MeanGlobal[X])
            MeanCorrectedBadVectors.append(NewVector)
        ############################################################
        # Compute covariance matrices:
        CovarArrayGood = self.GetCovarianceArray(MeanCorrectedGoodVectors)
        if VerboseFlag:
            print "CovarArrayGood:", CovarArrayGood
        CovarArrayBad = self.GetCovarianceArray(MeanCorrectedBadVectors)
        if VerboseFlag:
            print "CovarArrayBad:", CovarArrayBad
        # CovarArrayFull is the pooled within-group covariance matrix, it's
        # computed componentwise as weighted sum of CovarArrayGood and CovarArrayBad.
        CovarArrayFull = zeros((self.Size, self.Size), FloatType)
        for X in range(self.Size):
            for Y in range(self.Size):
                CovarArrayFull[X][Y] += CovarArrayGood[X][Y] * FeatureSet.TrueCount / float(FeatureSet.Count)
                CovarArrayFull[X][Y] += CovarArrayBad[X][Y] * FeatureSet.FalseCount / float(FeatureSet.Count)
        if VerboseFlag:
            print "CovarArrayFull:", CovarArrayFull
        ############################################################
        # Invert the covariance array:
        try:
            self.CI = InvertMatrix(CovarArrayFull)
        except:
            traceback.print_exc()
            print "Unable to invert covariance matrix!  Invalid feature set."
            return 0
        if VerboseFlag:
            print "CI:", self.CI
        self.GoodMuC = MatrixMultiply(self.CI, self.MeanGood)
        if VerboseFlag:
            print "GoodMuC:", self.GoodMuC
        self.BadMuC = MatrixMultiply(self.CI, self.MeanBad)
        if VerboseFlag:
            print "BadMuC:", self.BadMuC
        self.ConstantGood = -MatrixMultiply(self.MeanGood, self.GoodMuC) / 2.0
        self.ConstantBad = -MatrixMultiply(self.MeanBad, self.BadMuC) / 2.0
        #if VerboseFlag:
        print "LDA Constant good %.4f constant bad %.4f"%(self.ConstantGood, self.ConstantBad)

class SVMModel(LearnerClass):
    def __init__(self, FeatureList = None):
        self.Scaling = None
        self.SupportVectors = None
        self.PySVMReadyFlag = 0
        LearnerClass.__init__(self, FeatureList)
    def WriteSVMFeaturesToFile(self, FilePath, FeatureSet, ForceEqualCounts = 0):
        #print "TRUE vectors %s, FALSE vectors %s"%(len(FeatureSet.TrueVectors), len(FeatureSet.FalseVectors))
        #print "TRUE count %s, FALSE count %s"%(FeatureSet.TrueCount, FeatureSet.FalseCount)
        File = open(FilePath, "wb")
        if ForceEqualCounts:
            # Shuffle the tuples:
            TrueVectors = FeatureSet.TrueVectors[:]
            random.shuffle(TrueVectors)
            FalseVectors = FeatureSet.FalseVectors[:]
            random.shuffle(FalseVectors)
            # Try writing out equal numbers of true and false tuples:
            MaxIndex = min(FeatureSet.TrueCount, FeatureSet.FalseCount)
            MaxIndex = min(MaxIndex, MaxSVMFeatureCount)
            TrueVectors = TrueVectors[:MaxIndex]
            FalseVectors = FalseVectors[:MaxIndex]
        else:
            TrueVectors = FeatureSet.TrueVectors
            FalseVectors = FeatureSet.FalseVectors
        for Vector in TrueVectors:
            Str = "+1 "
            for FeatureIndex in range(len(self.FeatureList)):
                Str += "%d:%.8f "%(FeatureIndex + 1, Vector.Features[self.FeatureList[FeatureIndex]])
            File.write(Str + "\n")
        for Vector in FalseVectors:
            Str = "-1 "
            for FeatureIndex in range(len(self.FeatureList)):
                Str += "%d:%.8f "%(FeatureIndex + 1, Vector.Features[self.FeatureList[FeatureIndex]])
            File.write(Str + "\n")
        File.close()
    def Train(self, FeatureSet, VerboseFlag = 0):
        print "TRAINSVM()...", FeatureSet
        TempFeaturesFileName = "PTMFeatures.SVM.txt"
        TempScalingFileName = "PTMFeaturesSVMScale.txt"
        TempScaledFeaturesFileName = "PTMFeatures.SVMScaled.txt"
        TempModelFileName = "PTMFeatures.SVMScaled.txt.model"        
        # Write feature vectors, forcing equal true and false instance-counts:
        self.WriteSVMFeaturesToFile(TempFeaturesFileName, FeatureSet, 1)
        ###############################################################
        # SCALE the features, and remember the scaling:
        Command = r"%s -s %s %s > %s"%(PATH_SVMSCALE, TempScalingFileName, TempFeaturesFileName, TempScaledFeaturesFileName)
        print Command
        os.system(Command)
        # Read the scaling limits, for later use:
        File = open(TempScalingFileName, "rb")
        self.Scaling = File.read()
        File.close()
        os.remove(TempScalingFileName)
        print "Train!"
        ###############################################################
        # TRAIN the model.  We won't use cross-validation here, because in the future there'll be
        # a testing-set.
        Command = r"%s %s"%(PATH_SVMTRAIN, TempScaledFeaturesFileName)
        print Command
        os.system(Command)
        File = open(TempModelFileName, "rb")
        self.SupportVectors = File.read()
        File.close()
        ###############################################
        # Clean up temp-files:
        os.remove(TempFeaturesFileName)
        os.remove(TempScaledFeaturesFileName)
        os.remove(TempModelFileName)
    def Test(self, FeatureSet):
        if not self.Scaling or not self.SupportVectors:
            print "Error in SVMModel.Test(): We haven't trained (or loaded) yet!"
            return
        TempFeaturesFileName = "TestFeatures.SVM.txt"
        TempScalingFileName = "PTMFeaturesSVMScale.txt"
        TempScaledFeaturesFileName = "TestFeatures.SVMScaled.txt"
        TempModelFileName = "SVM.model"
        TempOutputFileName = "SVMPrediction.txt"
        TrueFlags = []
        for Tuple in FeatureSet.TrueVectors:
            TrueFlags.append(1)
        for Tuple in FeatureSet.FalseVectors:
            TrueFlags.append(0)        
        ########################################################################
        # WRITE the testing set to file:
        self.WriteSVMFeaturesToFile(TempFeaturesFileName, FeatureSet)
        # Write our scaling-info and our model to files:
        File = open(TempScalingFileName, "wb")
        File.write(self.Scaling)
        File.close()
        File = open(TempModelFileName, "wb")
        File.write(self.SupportVectors)
        File.close()
        # SCALE the testing set:
        Command = r"%s -r %s %s > %s"%(PATH_SVMSCALE, TempScalingFileName, TempFeaturesFileName, TempScaledFeaturesFileName)
        print Command
        os.system(Command)
        os.remove(TempFeaturesFileName)
        os.remove(TempScalingFileName)
        # Ok, now let's run svmpredict on all the instances in the TESTING set:
##        Command = r"%s %s %s %s"%(PATH_SVMPREDICT, TempScaledFeaturesFileName, TempModelFileName, TempOutputFileName)
##        print Command
##        os.system(Command)
        RunPySVM.Predict(TempScaledFeaturesFileName, TempModelFileName, TempOutputFileName)
        # Now read in the results, and assign scores to the vectors of the set:
        File = open(TempOutputFileName, "rb")
        InstanceIndex = 0
        TrueIndex = 0
        FalseIndex = 0
        SortedList = []
        for FileLine in File.xreadlines():
            Score = float(FileLine)
            if TrueFlags[InstanceIndex]:
                Vector = FeatureSet.TrueVectors[TrueIndex]
                TrueIndex += 1
            else:
                Vector = FeatureSet.FalseVectors[FalseIndex]
                FalseIndex += 1
            Vector.Score = float(FileLine)
            InstanceIndex += 1
        File.close()
        ########################################################################
        # Clean up temp-files:
        os.remove(TempScaledFeaturesFileName)
        os.remove(TempModelFileName)
        os.remove(TempOutputFileName)
    def SaveTextModel(self, Stub):
        ModelPath = "%s.model"%Stub
        File = open(ModelPath, "wb")
        File.write(self.SupportVectors)
        File.close()
        ScalingPath = "%s.range"%Stub
        File = open(ScalingPath, "wb")
        File.write(self.Scaling)
        File.close()
    def SaveModel(self, FileName):
        File = open(FileName, "wb")
        cPickle.dump("SVM", File)
        cPickle.dump(self.FeatureList, File)
        cPickle.dump(self.PValue, File)
        cPickle.dump(self.Scaling, File)
        cPickle.dump(self.SupportVectors, File)
        self.MixtureModel.PickleSelf(File)
        File.close()
    def LoadModel(self, FileName):
        File = open(FileName, "rb")
        cPickle.load(File) # model type
        self.FeatureList = cPickle.load(File)
        self.PValue = cPickle.load(File)
        self.Scaling = cPickle.load(File)
        self.SupportVectors = cPickle.load(File)
        self.MixtureModel = UnpickleMixtureModel(File)
        File.close()
    def PreparePySVM(self):
        """
        Prepare PySVM to score some features using our model.
        """
        # Support vectors:
        TempModelFileName = "TempModel.txt"
        File = open(TempModelFileName, "wb")
        File.write(self.SupportVectors)
        File.close()
        PySVM.LoadModel(TempModelFileName)
        os.remove(TempModelFileName)
        # Feature ranges:
        TempScalingFileName = "TempScaling.txt"
        File = open(TempScalingFileName, "wb")
        File.write(self.Scaling)
        File.close()
        PySVM.LoadScaling(TempScalingFileName)
        os.remove(TempScalingFileName)
        # And now, we can score many vectors quickly!
        self.PySVMReadyFlag = 1
    def ScoreInstance(self, Features):
        """
        Compute the score for this instance. 
        """
        if not self.PySVMReadyFlag:
            self.PreparePySVM()
        Vector = []
        for FeatureIndex in self.FeatureList:
            Vector.append(Features[FeatureIndex])
        Score = PySVM.ScaleAndScore(Vector)
        return Score
    
class LogitModel(LearnerClass):
    """
    A maximum-likelihood logistic regression model.  The model's parameters
    are tuned using the Newton-Raphson algorithm.  See p98 in Hastie/Tibshirani,
    The Elements of Statistical Learning.  
    """
    def GetMMBinMultiplier(self):
        return 40.0 #default
    def ComputePValues(self, SortedList):
        pass # The score we output IS a probability!
    def GetFixedTuples(self, FeatureSet):
        # Return fixed-up tuples.  Keep a random selection of
        # true and of false tuples.  
        VectorSize = len(self.FeatureList) + 1 # add one for the CONSTANT input
        AllTuples = [] # entries (TrueFlag, FeatureTuple)
        KeepCount = min(FeatureSet.TrueCount, FeatureSet.FalseCount, 500)
        random.shuffle(FeatureSet.TrueVectors)
        random.shuffle(FeatureSet.FalseVectors)
        for Vector in FeatureSet.TrueVectors[:KeepCount]:
            FixedTuple = [1.0]
            for FeatureIndex in self.FeatureList:
                FixedTuple.append(Vector.Features[FeatureIndex])
            AllTuples.append((1, tuple(FixedTuple)))
        for Vector in FeatureSet.FalseVectors[:KeepCount]:
            FixedTuple = [1.0]
            for FeatureIndex in self.FeatureList:
                FixedTuple.append(Vector.Features[FeatureIndex])
            AllTuples.append((0, tuple(FixedTuple)))        
        return AllTuples
    def Train(self, FeatureSet, VerboseFlag = 0):
        VectorSize = len(self.FeatureList) + 1 # add one for the CONSTANT input
        AllTuples = self.GetFixedTuples(FeatureSet)
        random.shuffle(AllTuples)
        #################################################################
        # Train the model - set the weight-vector self.Beta
        # Initialize vector self.Beta, all zeroes:
        self.Beta = zeros(VectorSize)
        TupleCount = len(AllTuples)
        # Initialize vector Y, indicating which vectors are true:
        Y = zeros(TupleCount, FloatType)
        for I in range(TupleCount):
            if AllTuples[I][0]:
                Y[I] = 1.0
            else:
                Y[I] = 0.0
        # Initialize the input matrix X:
        X = zeros((TupleCount, VectorSize), FloatType)
        for I in range(TupleCount):
            #X[I][0] = 1.0
            for J in range(VectorSize):
                X[I][J] = AllTuples[I][1][J]
        XT = transpose(X)
        PrevLogLikelihood = None
        CycleCount = 0
        while 1:
            # Compute the current log-likelihood:
            LogLikelihood = 0
            for I in range(TupleCount):
                BX = MatrixMultiply(self.Beta, X[I])
                LogLikelihood += Y[I] * BX
                LogLikelihood -= math.log(1 + math.exp(BX))
            if PrevLogLikelihood != None:
                if VerboseFlag:
                    print "Log likelihood: %s (prev %s)"%(LogLikelihood, PrevLogLikelihood)
                Improvement = PrevLogLikelihood - LogLikelihood
                if Improvement < 0.001 and Improvement >= 0:
                    print "Reached optimum: Stop now!"
                    break
            PrevLogLikelihood = LogLikelihood
            # Compute the vector P:
            P = zeros(TupleCount, FloatType)
            for I in range(TupleCount):
                self.BetaSum = 0
                Tuple = AllTuples[I][1]
                for J in range(VectorSize):
                    self.BetaSum += self.Beta[J] * Tuple[J]
                Exp = math.exp(self.BetaSum)
                P[I] = Exp / (1.0 + Exp)
            # Compute the diagonal matrix W:
            W = zeros((TupleCount, TupleCount), FloatType)
            for I in range(TupleCount):
                W[I][I] = P[I] * (1.0 - P[I])
            try:
                WI = numpy.linalg.inv(W)
            except:
                traceback.print_exc()
                print "** Error: Unable to perform logistic regression due to singular matrix."
                print "Feature list was:", self.FeatureList
                return None
            # Compute the "response vector" z:
            z = MatrixMultiply(X, self.Beta)
            Diff = Y - P
            z += MatrixMultiply(WI, Diff)
            # Compute the new self.Beta:
            Product = MatrixMultiply(XT, W)
            Product = MatrixMultiply(Product, X)
            ProdI = numpy.linalg.inv(Product)
            Product = MatrixMultiply(ProdI, XT)
            Product = MatrixMultiply(Product, W)
            NewBeta = MatrixMultiply(Product, z)
            if VerboseFlag:
                print "Old self.Beta:", self.Beta
                print "New self.Beta:", NewBeta
            self.Beta = NewBeta
            CycleCount += 1
            if CycleCount >= 100:
                print "100 cycles performed; stopping now!"
                break
    def ScoreInstance(self, Features):
        FixedFeatures = [1.0,]
        for FeatureIndex in self.FeatureList:
            FixedFeatures.append(Features[FeatureIndex])
        BX = 0
        for I in range(len(FixedFeatures)):
            BX += self.Beta[I] * FixedFeatures[I]
        try:
            Exp = math.exp(BX)
        except:
            print "** exponent unreachable:", BX
            print "Features:", FixedFeatures
            raise ValueError, "Features out-of-range!"
        Score = Exp / (1.0 + Exp)
        return Score
    def Test(self, FeatureSet):
        for Vector in FeatureSet.AllVectors:
            FixedFeatures = [1.0,]
            for FeatureIndex in self.FeatureList:
                FixedFeatures.append(Vector.Features[FeatureIndex])
            BX = 0
            for I in range(len(FixedFeatures)):
                BX += self.Beta[I] * FixedFeatures[I]
            Exp = math.exp(BX)
            Vector.Score = Exp / (1.0 + Exp)
    def SaveModel(self, FileName):
        File = open(FileName, "wb")
        cPickle.dump("LOGIT", File)
        cPickle.dump(self.FeatureList, File)
        cPickle.dump(self.PValue, File)
        cPickle.dump(self.Beta, File)
        File.close()
    def LoadModel(self, FileName):
        File = open(FileName, "rb")
        cPickle.load(File) # model type
        self.FeatureList = cPickle.load(File)
        self.PValue = cPickle.load(File)
        self.Beta = cPickle.load(File)
        File.close()
   
def Test():
    pass

if __name__ == "__main__":
    # Command-line invocation: Test model loading/saving
    Test()



#Title:          ResultsParser.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.

"""
Constants and methods for parsing (Inspect) search results
"""
import os
import random
class Columns:

    DefaultInspectHeader = "#SpectrumFile\tScan#\tAnnotation\tProtein\tCharge\tMQScore\tLength\tTotalPRMScore\tMedianPRMScore\tFractionY\tFractionB\tIntensity\tNTT\tInspectFDR\tF-Score\tDeltaScore\tDeltaScoreOther\tRecordNumber\tDBFilePos\tSpecFilePos\tPrecursorMZ\tPrecursorMZError\tSpecIndex"


    def __init__(self):
        self.header = self.initializeHeaders(self.DefaultInspectHeader)

    def initializeHeaders(self,Header):
        if Header[0] == '#':
            Header = Header[1:]

        self.headers = Header.lower().split("\t")
        
    def getIndex(self,headerVal):
        
        for i in range(0,len(self.headers)):
            if headerVal.lower() == self.headers[i]:
                return i

        return -1
        

#    "Constants for which columns contain which data"
#    SpectrumFile = 0
#    ScanNumber = 1
#    Annotation = 2
#    ProteinName = 3
#    Charge = 4
#    MQScore = 5
#    Length = 6
#    NTT = 12
#    PValue = 13
#    FScore = 14
#    DeltaScoreAny = 15
#    DeltaScore = 16
#    ProteinID = 17
#    DBPos = 18
#    FileOffset = 19 #Spectrum File pos
#    ParentMZ = 20 #Corrected, associated with tweak
#    MZError = 21

#    #More columns for splicing
#    Chromosome = 22
#    Strand = 23
#    GenomicPost = 24
#    SplicedSequence = 25
#    Splices = 26
#    SearchDB = 27



class SpectrumOracleMixin:
    def __init__(self):
        self.SpectrumOracle = {}
    def FixSpectrumPath(self, Path):
        FileName = os.path.split(Path)[-1]
        Stub = os.path.splitext(FileName)[0]
        return self.SpectrumOracle.get(Stub, Path)
    def PopulateSpectrumOracle(self, RootDirectory):
        """
        Used when mzxml files are spread over multiple subdirectories.
        MZXMLOracle[Stub] = full path to the corresponding MZXML file
        Used with -M option (not with -s option)
        """
        if not RootDirectory or not os.path.exists(RootDirectory):
            return
        print "Populate oracle from %s..."%RootDirectory
        for SubFileName in os.listdir(RootDirectory):
            # Avoid expensive iteration through results directories:
            if SubFileName[:7] == "Results":
                continue
            SubFilePath = os.path.join(RootDirectory, SubFileName)
            if os.path.isdir(SubFilePath):
                self.PopulateSpectrumOracle(SubFilePath)
                continue
            (Stub, Extension) = os.path.splitext(SubFileName)
            Extension = Extension.lower()
            if Extension == ".mzxml":
                self.SpectrumOracle[Stub] = os.path.join(RootDirectory, SubFileName)
            elif Extension == ".mgf":
                self.SpectrumOracle[Stub] = os.path.join(RootDirectory, SubFileName)
            elif Extension == ".ms2":
                self.SpectrumOracle[Stub] = os.path.join(RootDirectory, SubFileName)
                
class ResultsParser:
    def __init__(self, *args, **kw):
        #self.Columns = Columns
        self.Running = 1
    def ProcessResultsFiles(self, FilePath, Callback, MaxFilesToParse = None, QuietFlag = 0):
        """
        Function for applying a Callback function to one search-reuslts file, or to every
        search-results file in a directory.
        """
        print "ResultsParser:%s"%FilePath
        FileCount = 0
        if os.path.isdir(FilePath):
            FileNames = os.listdir(FilePath)
            random.shuffle(FileNames)
            for FileNameIndex in range(len(FileNames)):
                FileName = FileNames[FileNameIndex]
                if not QuietFlag:
                    print "(%s/%s) %s"%(FileNameIndex, len(FileNames), FileName)
                (Stub, Extension) = os.path.splitext(FileName)
                if Extension.lower() not in (".txt", ".filtered", ".res", ".csv", ".out"):
                    continue
                FileCount += 1
                SubFilePath = os.path.join(FilePath, FileName)
                apply(Callback, (SubFilePath,))
                # Don't parse every single file, that will take too long!
                if MaxFilesToParse != None and FileCount > MaxFilesToParse:
                    break 
        else:
            apply(Callback, (FilePath,))
    



#Title:          Score.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.


"""
Score.py: Simple wrapper for inspect scoring
"""
import sys
import string
import PyInspect

def FormatTuple(Tuple):
    Str = "("
    for Entry in Tuple:
        Str += "%.4g, "%Entry
    Str = Str[:-2]
    Str += ")"
    return Str

ColonBits = sys.argv[1].split(":")
try:
    FileOffset = int(ColonBits[-1])
    FileName = string.join(ColonBits[:-1], ":")
except:
    FileName = sys.argv[1]
    FileOffset = 0
    
Spectrum = PyInspect.Spectrum(FileName, FileOffset)
#Result = Spectrum.ScorePeptideDetailed(sys.argv[2])
Result = Spectrum.ScorePeptideDetailed(sys.argv[2])
Str = "MQ %.4f %s"%(Result[0], FormatTuple(Result[1:]))
print Str
